/*
 * bcm_rdpa.c - Broadcom RDPA QoS library implementation
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: oskar.viljasaar@iopsys.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include <bcmnet.h>
#include <bcmswapitypes.h>
#include <rdpa_types.h>
#include <rdpa_drv.h>

#include "qos.h"

#define IFNAME_SIZE 16
#define RDPACTL_IOCTL_FILE_NAME "/dev/bcmrdpa"

#define get_dir(dev_type, rdpa_if) \
  ((((dev_type == RDPA_IOCTL_DEV_PORT) && (!rdpa_if_is_wan(rdpa_if))) ||\
    (dev_type == RDPA_IOCTL_DEV_NONE))? rdpa_dir_ds : rdpa_dir_us)

/**
 *  Initialize the socket to bcmsw interface
 *  @param p_ifr output parameter pointer to ifreq structure
 *  return : integer +ve socket file discriptor on success and -1 for failure.
 */
static inline int ethswctl_init(struct ifreq *p_ifr)
{
	int skfd;
	char wan[IFNAME_SIZE] = {0};

	/* Open a basic socket */
	if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		syslog(LOG_ERR, "socket open error\n");
		return -1;
	}

	/* Get the name -> if_index mapping for ethswctl */
	strncpy(p_ifr->ifr_name, "bcmsw", strlen("bcmsw"));
	if (ioctl(skfd, SIOCGIFINDEX, p_ifr) < 0 ) {
		// As per bcmkernel eth0 seems to be reserved for bcmsw failsafe
		strncpy(p_ifr->ifr_name, "eth0", strlen("eth0"));
		if (ioctl(skfd, SIOCGIFINDEX, p_ifr) < 0 ) {
			close(skfd);
			syslog(LOG_ERR, "neither bcmsw nor eth0 exist\n");
			return -1;
		}
	}

	return skfd;
}

/**
 *  To get rdpa interface form linux interface name like eth0 etc.
 *  @param ifname input parameter pointer to linux interface name
 *  @param rdpa_if_p output parameter pointer to rdpa interface
 *  return : integer +ve socket file discriptor on success and -1 for failure.
 */
static int bcm_enet_get_rdpa_if_from_ifname(const char* ifname, int* rdpa_if_p)
{
	int skfd, err = 0;
	struct ifreq ifr;
	struct ethswctl_data ifdata = {0};
	struct ethswctl_data *e = &ifdata;

	if ((skfd = ethswctl_init(&ifr)) < 0) {
		syslog(LOG_ERR, "ethswctl_init failed. \n");
		return skfd;
	}
	ifr.ifr_data = (char *)&ifdata;

	e->op = ETHSWRDPAPORTGETFROMNAME;
	e->type = TYPE_GET;
	strncpy(e->ifname, ifname, strlen(ifname));

	if((err = ioctl(skfd, SIOCETHSWCTLOPS, &ifr))) {
		syslog(LOG_ERR, "ioctl command return error %d\n", err);
		close(skfd);
		return err;
	} else {
		syslog(LOG_ERR, "ifname - %s rdpa_if %d\n", e->ifname, e->val);
		close(skfd);
		*rdpa_if_p = e->val;
		return err;
	}
}

/**
 *  Actual ioctl call for rdpa file /dev/bcmrdpa
 *  @param code input parameter to RDPA_IOC_TM
 *  @param ctx output parameter for ioctl
 *  return : integer 0 on success and -1 on failure.
 */
static inline int rdpa_ioctl(int code, uintptr_t ctx)
{
	int fd, ret = 0;
	fd = open(RDPACTL_IOCTL_FILE_NAME, O_RDWR);
	if (fd < 0) {
		syslog(LOG_ERR, "Open system call failed for : %s : %d\n", RDPACTL_IOCTL_FILE_NAME, fd);
		return -1;
	}

	ret = ioctl(fd, code, ctx);
	if (ret)
		syslog(LOG_ERR, "ioctl failed, ret=%d\n", ret);

	close(fd);
	return ret;
}

/**
 * Fill rdpa structure and call ioctl to get stat
 *  @param dev_type input parameter for device type
 *  @param dev_id input parameter for device id
 *  @param dir input parameter for traffic direction
 *  @param qid input parameter for queue id
 *  @param pStats output parameter pointer to rdpa_stat_1way_t
 *  return : integer 0 on success and -1 on failure.
 */
static int rdpa_ctl_get_queue_stats( int dev_type, int dev_id, int dir,
				  int qid, rdpa_stat_1way_t *pStats)
{
	int rc = 0;
	rdpa_drv_ioctl_tm_t tm;

	memset(&tm, 0, sizeof(rdpa_drv_ioctl_tm_t));

	tm.cmd = RDPA_IOCTL_TM_CMD_GET_QUEUE_STATS;
	tm.dev_type = dev_type;
	tm.dev_id = dev_id;
	tm.dir = dir;
	tm.q_id = qid;

	rc = rdpa_ioctl(RDPA_IOC_TM, (uintptr_t)&tm);
	if (!rc) {
		*pStats = tm.qstats;

		syslog(LOG_ERR, "[Stats] PASSED: packets=%u bytes=%u  DISCARDED: packets=%u bytes=%u\n",
		       pStats->passed.packets, pStats->passed.bytes,
		       pStats->discarded.packets, pStats->discarded.bytes);
	}

	return rc;
}

/**
 *  Final queue stats to be filled in qos_stat structure from rdpa_get_queue_stats
 *  @param dev_type input parameter for device type
 *  @param rdpa_if input parameter for actual rdpa interface
 *  @param qid input parameter for queue id
 *  @param qstats output parameter pointer to qos_stats
 *  return : integer 0 on success and -1 on failure.
 */
static int get_bcm_queue_stats(int dev_type, int rdpa_if, int qid,
			      struct qos_stats *qstats)
{
	int ret = 0;

	int  rc;
	rdpa_traffic_dir dir;
	rdpa_stat_1way_t stats;

	dir = get_dir(dev_type, rdpa_if);

	if ((rc = rdpa_ctl_get_queue_stats(dev_type, rdpa_if, dir, qid, &stats))) {
		syslog(LOG_ERR, "rdpa_ctl_ge_queu_stats ERROR! rdpa_if=%d qid=%d rc=%d\n",
		       rdpa_if, qid, rc);
		return -1;
	}

	qstats->tx_packets = stats.passed.packets;
	qstats->tx_bytes = stats.passed.bytes;
	qstats->tx_dropped_packets = stats.discarded.packets;
	qstats->tx_dropped_bytes = stats.discarded.bytes;

	return ret;
}

/**
 *  To get the queue stats for an RDPA device:
   - determine the RDPA device ID through the bcmsw device
   - Fill up a rdpa_drv_ioctl_tm_t struct with the request and the RDPA
     device ID, pass it through an ioctl to /dev/bcmrdpa
 *  @param ifname input parameter for linux interface
 *  @param qid input parameter for queue id
 *  @param qstats output parameter pointer to qos_stats
 */
static int bcm_get_stats(const char *ifname,
			 int queue_id, struct qos_stats *stats)
{
	int ret;
	int rdpa_if;

	ret = bcm_enet_get_rdpa_if_from_ifname(ifname, &rdpa_if);
	if (ret == -1) {
		syslog(LOG_ERR,"Cannot find rdpa_if\n");
		return ret;
	}

	ret = get_bcm_queue_stats(0, rdpa_if, queue_id, stats);
	if (ret == -1) {
		syslog(LOG_ERR,"Cannot find rdpa_if\n");
		return ret;
	}

	return ret;
}

const struct qos_ops bcm_rdpa_ops = {
	.ifname = "eth",
	.get_stats = bcm_get_stats,
};
