/*
 * wlctrl.c - broadcom related utilities
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <sys/types.h>
#include <linux/if_packet.h>
#include <net/if.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <net/if.h>
#include <time.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <endian.h>
#include <byteswap.h>
#include <sys/ioctl.h>
#include <wlc_types.h>

#include "util.h"
#include "nlwifi.h"
#include "wifi.h"
#include "debug.h"
#include "wlctrl.h"

#include "brcm_nl80211.h"
#include "wpactrl_util.h"

#define BCMSWAP16 bswap_16
#define BCMSWAP32 bswap_32
#define BCMSWAP64 bswap_64

struct wl_arg {
	const char *ifname;

	int cmd;		/* eg WLC_GET_VAR */
	const char *iovar;	/* optional */

	const void *param;	/* optional */
	int paramlen;

	void *buf;		/* output buffer */
	int buflen;

	bool set;
	int len;		/* return len */
};

static int f2c(int freq)
{
	int chan;

	if (freq > 5000)
		chan = (freq - 5000) / 5;
	else if (freq == 2484)
		chan = 14;
	else
		chan = (freq - 2407) / 5;

	return chan;
}

static int c2f(int chan)
{
	int freq = 0;

	if (chan >= 1 && chan <= 13)
		freq = 2407 + chan * 5;
	if (chan == 14)
		freq = 2484;
	if (chan >= 36)
		freq = 5000 + chan * 5;

	return freq;
}

static uint32_t cac_time(const char *cc, int chan)
{
	uint32_t cac_time = 60;

	if (!strcmp(cc, "US") || !strcmp(cc, "JP"))
		return cac_time;

	if (chan == 120 || chan == 124 || chan == 128)
		cac_time = 600;

	return cac_time;
}

static int wl_ioctl(struct wl_arg *arg)
{
	char buf[WLC_IOCTL_MAXLEN];
	struct ifreq ifr;
	wl_ioctl_t wioc;
	int len = 0;
	int ret = -1;
	int s;

	if (WARN_ON(arg->buflen > sizeof(buf)))
		goto exit;

	memset(buf, 0, sizeof(buf));
	memset(&wioc, 0, sizeof(wioc));
	memset(&ifr, 0, sizeof(ifr));

	wioc.cmd = arg->cmd;
	wioc.buf = buf;
	wioc.len = sizeof(buf);
	wioc.set = arg->set;

	if (arg->iovar) {
		len = strlen(arg->iovar) + 1;
		if (WARN_ON(len > sizeof(buf)))
			goto exit;
		strcpy(buf, arg->iovar);
		if (arg->cmd == WLC_SET_VAR)
			wioc.len = len + arg->paramlen;
	}

	if (arg->param) {
		if (WARN_ON((len + arg->paramlen) > sizeof(buf)))
			goto exit;
		memcpy(&buf[len], arg->param, arg->paramlen);
	}

	strncpy(ifr.ifr_name, arg->ifname, IFNAMSIZ);
	ifr.ifr_data = (caddr_t) &wioc;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (WARN_ON(s < 0))
		goto exit;

	fcntl(s, F_SETFD, fcntl(s, F_GETFD) | FD_CLOEXEC);

	ret = ioctl(s, SIOCDEVPRIVATE, &ifr);
	close(s);

	if (ret)
		goto exit;

	arg->len = wioc.len;

	if (arg->buf)
		memcpy(arg->buf, buf, arg->buflen);

exit:
	libwifi_dbg("[%s] %s cmd %d %s ret %d\n", arg->ifname, __func__, wioc.cmd, arg->iovar ?: "", ret);
	return ret;
}

static int wl_iovar_set(const char *ifname, char *iovar, void *param, int paramlen,
			void *buf, int buflen)
{
	struct wl_arg arg = {
		.ifname = ifname,
		.cmd = WLC_SET_VAR,
		.iovar = iovar,
		.param = param,
		.paramlen = paramlen,
		.buf = buf,
		.buflen = buflen
	};

	return wl_ioctl(&arg);
}

static int wl_iovar_get_noparam(const char *ifname, char *iovar, void *buf, int buflen)
{
	struct wl_arg arg = {
		.ifname = ifname,
		.cmd = WLC_GET_VAR,
		.iovar = iovar,
		.buf = buf,
		.buflen = buflen
	};

	return wl_ioctl(&arg);
}

static int wl_iovar_get(const char *ifname, char *iovar, void *param, int paramlen,
			void *buf, int buflen)
{
	struct wl_arg arg = {
		.ifname = ifname,
		.cmd = WLC_GET_VAR,
		.iovar = iovar,
		.param = param,
		.paramlen = paramlen,
		.buf = buf,
		.buflen = buflen
	};

	return wl_ioctl(&arg);
}

static int wl_swap(const char *ifname)
{
	unsigned int magic = 0;
	struct wl_arg arg = {
		.ifname = ifname,
		.cmd = WLC_GET_MAGIC,
		.buf = &magic,
		.buflen = sizeof(magic)
	};

	if (WARN_ON(wl_ioctl(&arg) < 0))
		return 0; /* don't swap on error */

	/* is endian swap needed */
	if (magic == WLC_IOCTL_MAGIC)
		return 0;
	else if (bswap_32(magic) == WLC_IOCTL_MAGIC)
		return 1;
	else
		libwifi_err("[%s] wrong magic, driver 0x%x vs header 0x%x\n", ifname, magic, WLC_IOCTL_MAGIC);

	return 0;
}

static uint16_t wl_swap_16(const char *ifname, uint16_t val)
{
	return wl_swap(ifname) ? bswap_16(val) : val;
}

static uint32_t wl_swap_32(const char *ifname, uint32_t val)
{
	return wl_swap(ifname) ? bswap_32(val) : val;
}

static uint64_t wl_swap_64(const char *ifname, uint64_t val)
{
	return wl_swap(ifname) ? bswap_64(val) : val;
}

static int wl_vendor(struct wl_arg *arg)
{
	struct bcm_nlmsg_hdr *req;
	uint8_t out[WLC_IOCTL_MAXLEN];
	uint8_t in[2048];
	int outlen = 0;
	int inlen = 0;
	int len = 0;
	int ret;

	memset(out, 0, sizeof(out));
	memset(in, 0, sizeof(in));

	req = (struct bcm_nlmsg_hdr *) in;
	inlen = sizeof(*req);

	req->cmd = arg->cmd;
	req->len = arg->buflen;

	req->offset = sizeof(*req);
	req->set = arg->set;

	if (arg->iovar) {
		len = strlen(arg->iovar) + 1;
		if (WARN_ON((len + sizeof(*req)) > sizeof(in)))
			return -1;
		strcpy((char *)&in[req->offset], arg->iovar);
		inlen += len;
	}

	if (arg->param) {
		if (WARN_ON((len + arg->paramlen + sizeof(*req)) > sizeof(in)))
			return -1;
		memcpy(&in[len + req->offset], arg->param, arg->paramlen);
		inlen += arg->paramlen;
	}

	ret = nlwifi_vendor_cmd(arg->ifname, OUI_BRCM, BRCM_VENDOR_SCMD_PRIV_STR,
				in, inlen, out, &outlen);

	libwifi_dbg("[%s] %s called cmd %d ret %d and len %d\n", arg->ifname, __func__, arg->cmd, ret, outlen);

	arg->len = outlen;

	if (arg->buf)
		memcpy(arg->buf, out, arg->buflen);

	return ret;
}

static int wl_ctl(struct wl_arg *arg)
{
	int ret;

	/* Today a lot of vendor commands fail - prefer ioctl */
	return wl_ioctl(arg);

	ret = wl_vendor(arg);
	if (ret)
		ret = wl_ioctl(arg);

	return ret;
}

static const uint8_t wf_chspec_bw_mhz[] = {5, 10, 20, 40, 80, 160, 160};
static const uint8 wf_5g_80m_chans[] = {42, 58, 106, 122, 138, 155};
#define WF_NUM_BW (sizeof(wf_chspec_bw_mhz)/sizeof(uint8_t))

static uint8_t chanspec_to_ctrlchannel(chanspec_t c)
{
	uint8_t bw;
	uint8_t center_ch;
	uint8_t sb;

	if (CHSPEC_BW_LE20(c))
		return CHSPEC_CHANNEL(c);

	sb = CHSPEC_CTL_SB(c) >> WL_CHANSPEC_CTL_SB_SHIFT;
	if (!CHSPEC_IS8080(c)) {
		bw = (c & WL_CHANSPEC_BW_MASK) >>
				WL_CHANSPEC_BW_SHIFT;
		bw = bw >= WF_NUM_BW ? 0 :
			wf_chspec_bw_mhz[bw];
		center_ch = CHSPEC_CHANNEL(c) >>
				WL_CHANSPEC_CHAN_SHIFT;
	} else {
		uint8_t chan_id = CHSPEC_CHANNEL(c);

		bw = 80;
		center_ch = wf_5g_80m_chans[chan_id];
	}
	return ((center_ch - (((bw - 20) / 2) / 5)) + sb * 4);
}

static int wl_ioctl_iface_get_beacon(const char *name, uint8_t *bcn, size_t *len)
{
	char buf[WLC_IOCTL_MAXLEN] = { 0 };
	int ret;
	size_t bcn_len;

	if (WARN_ON(!bcn) || WARN_ON(!len) || WARN_ON(*len == 0))
		return -EINVAL;

	memset(bcn, 0, *len);

	ret = wl_iovar_get_noparam(name, "beacon_info", (void*)buf, WLC_IOCTL_MAXLEN);

	if (WARN_ON(ret))
		return ret;

	bcn_len = (size_t)wl_swap_32(name, *(uint32_t*)buf);

	if (WARN_ON(bcn_len <= 0 || *len < bcn_len )) {
		*len = 0;
		return -1;
	}

	memcpy(bcn, buf + 4, bcn_len);
	*len = bcn_len;

	return 0;
}

static int wl_ioctl_get_version(const char *name, char *version, size_t *len)
{
	char buf[256] = { 0 };
	const char *seps = "\t\n\v\f\r";
	int ret;
	int i;

	memset(version, 0, *len);

	ret = wl_iovar_get_noparam(name, "ver", (void*)buf, 256);
	if (WARN_ON(ret))
		return ret;

	i = strlen(buf) - 1;
	while (i > 0 && strchr(seps, buf[i]) != NULL) {
		buf[i] = '\0';
		i--;
	}
	if (*len > strlen(buf)) {
		*len = strlen(buf);
		strcpy(version, buf);
	} else {
		*len -= 1;
		strncpy(version, buf, *len);
		version[*len] = '\0';
	}

	return 0;
}

static int wl_cli_get_noise(const char *name, int *noise)
{
	char buf[64] = {0};
	int val = 0;
	int status;

	status = Cmd(buf, sizeof(buf), "wl -i %s noise", name);

	if (WARN_ON(status))
		return status;

	if (WARN_ON(sscanf(buf, "%d", &val) != 1))
		return -1;

	*noise = val;
	return 0;
}

static int wl_cli_get_country(const char *name, char *alpha2)
{
	char buf[256] = {0};
	int status;

	status = Cmd(buf, sizeof(buf), "wl -i %s country", name);

	if (WARN_ON(status))
		return status;

	if (WARN_ON(snprintf(alpha2, 3, "%s", buf) < 2))
		return -1;

	return 0;
}

static int wl_cli_get_rssi(const char *name, int8_t *rssi)
{
	char buf[64] = {0};
	int status;

	status = Cmd(buf, sizeof(buf), "wl -i %s rssi", name);
	if (WARN_ON(status))
		return status;

	if (WARN_ON(sscanf(buf, "%hhi", rssi) != 1))
		return -1;

	return 0;
}

static int wl_cli_get_macaddr(const char *name, uint8_t *macaddr)
{
	char buf[512] = {0};
	int status;

	status = Cmd(buf, sizeof(buf), "wl -i %s cur_etheraddr", name);
	if (WARN_ON(status))
		return status;

	status = sscanf(buf, "cur_etheraddr: %2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx",
			&macaddr[0], &macaddr[1], &macaddr[2],
			&macaddr[3], &macaddr[4], &macaddr[5]);

	if (WARN_ON(status != 6))
		return -1;

	return 0;
}

static int wl_cli_get_bssid(const char *name, uint8_t *bssid)
{
	char buf[512] = {0};
	int status;

	status = Cmd(buf, sizeof(buf), "wl -i %s bssid", name);
	if (WARN_ON(status))
		return status;

	status = sscanf(buf, "%2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx",
			&bssid[0], &bssid[1], &bssid[2],
			&bssid[3], &bssid[4], &bssid[5]);

	if (WARN_ON(status != 6))
		return -1;

	return 0;
}

int bcmwl_radio_get_noise(const char *name, int *noise)
{
	uint32_t val;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_PHY_NOISE,
		.buf = &val,
		.buflen = sizeof(val)
	};

	ret = wl_ioctl(&arg);
	if (!ret) {
		*noise = wl_swap_32(name, val);
		return ret;
	}

	libwifi_dbg("[%s] %s fallback to wl\n", name, __func__);
	ret = wl_cli_get_noise(name, noise);

	return ret;
}

static int wl_cli_get_chanspec(const char *name, chanspec_t *val)
{
	char buf[64] = {0};
	int status;

	status = Cmd(buf, sizeof(buf), "wl -i %s chanspec", name);

	if (WARN_ON(status))
		return status;

	if (WARN_ON(sscanf(strstr(buf,"(0x"), "(0x%x", (uint32_t *) val) != 1))
		return -1;

	return 0;
}

static int bcmwl_radio_get_chanspec(const char *name, chanspec_t *val)
{
	int ret;
	int swap = wl_swap(name);
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_VAR,
		.iovar = "chanspec",
		.buf = val,
		.buflen = sizeof(*val)
	};

	ret = wl_ctl(&arg);
	if (ret == 0)
		goto swap;

	libwifi_dbg("[%s] %s fallback to wl ret=%d\n", name, __func__, ret);
	ret = wl_cli_get_chanspec(name, val);

	if (WARN_ON(ret))
		return ret;

	goto end;

swap:
	*val = swap ? BCMSWAP16(*val) : *val;
end:
	return ret;
}

int bcmwl_radio_get_channel(const char *name, uint32_t *channel, enum wifi_bw *bw)
{
	chanspec_t val;
	int ret;

	ret = bcmwl_radio_get_chanspec(name, &val);
	if (WARN_ON(ret))
		return ret;

	*bw = BW_UNKNOWN;

	if (CHSPEC_IS20(val)) *bw = BW20;
	if (CHSPEC_IS40(val)) *bw = BW40;
	if (CHSPEC_IS80(val)) *bw = BW80;
	if (CHSPEC_IS160(val)) *bw = BW160;
	if (CHSPEC_IS8080(val)) *bw = BW8080;
	*channel = chanspec_to_ctrlchannel(val);

	return 0;
}

int bcmwl_radio_get_sideband(const char *name, enum wifi_chan_ext *sideband)
{
	chanspec_t val;
	int ret;

	ret = bcmwl_radio_get_chanspec(name, &val);
	if (WARN_ON(ret))
		return ret;

	*sideband= EXTCH_NONE;

	if(CHSPEC_SB_UPPER(val)) *sideband = EXTCH_ABOVE;
	if(CHSPEC_SB_LOWER(val)) *sideband = EXTCH_BELOW;

	return 0;
}

int bcmwl_radio_get_country(const char *name, char *alpha2)
{
	char val[3];
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_COUNTRY,
		.buf = &val,
		.buflen = sizeof(val)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);

	ret = wl_ctl(&arg);
	if (!ret) {
		snprintf(alpha2, 3, "%s", val);
		return ret;
	}

	libwifi_dbg("[%s] %s fallback to wl\n", name, __func__);
	ret = wl_cli_get_country(name, alpha2);

	return ret;
}

/* Best if we could include this directly from SDK */
#define ALIGN_SIZE(size, boundary) (((size) + (boundary) - 1) \
                                                 & ~((boundary) - 1))
typedef struct bcm_xtlv {
        uint16  id;
        uint16  len;
        uint8   data[1];
} bcm_xtlv_t;


uint8_t *bcmwl_get_tlv(int swap, uint8_t *buf, size_t len, uint16_t id)
{
	bcm_xtlv_t *xtlv;
	uint8_t *ptr;
	uint8_t *res;

	ptr = buf;
	res = NULL;

	while ((ptr - buf) < len) {
		xtlv = (bcm_xtlv_t *) ptr;
		if (swap) {
			xtlv->id = BCMSWAP16(xtlv->id);
			xtlv->len = BCMSWAP16(xtlv->len);
		}

		libwifi_dbg("tlv id 0x%x len %u\n", xtlv->id, xtlv->len);
		if (xtlv->id == id) {
			res = xtlv->data;
			break;
		}
		ptr += ALIGN_SIZE(sizeof(xtlv->id) + sizeof(xtlv->len) + xtlv->len, 4);
	}

	return res;
}

int bcmwl_radio_get_stats(const char *name, struct wifi_radio_stats *s)
{
	char buf[WLC_IOCTL_MAXLEN] = {};
	wl_cnt_info_t *c = (wl_cnt_info_t *) buf;
	wl_cnt_wlc_t *cnt;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_VAR,
		.iovar = "counters",
		.buf = buf,
		.buflen = sizeof(buf),
	};
	int swap;
	int ret;

	libwifi_dbg("[%s] %s called\n", name, __func__);
	memset(s, 0, sizeof(*s));

	swap = wl_swap(name);
	ret = wl_ctl(&arg);
	if (ret) {
		libwifi_err("[%s] %s error %d\n", name, __func__, ret);
		return -1;
	}

	/* TODO check headers */
	c = (wl_cnt_info_t *) buf;
	if (swap) {
		c->version = BCMSWAP16(c->version);
		c->datalen = BCMSWAP16(c->datalen);
	}

	if (c->version == 11 || c->version == 10)
		cnt = (wl_cnt_wlc_t *)c->data;
	else if (c->version == 30)
		cnt = (wl_cnt_wlc_t *) bcmwl_get_tlv(swap, c->data, c->datalen, WL_CNT_XTLV_WLC);  /* xtlv formatted */
	else {
		libwifi_err("'counters' version not supported!\n");
		return 0;
	}

	if (WARN_ON(!cnt))
		return -1;

	if (swap) {
		s->tx_pkts = (unsigned long)BCMSWAP32(cnt->txframe);
		s->tx_bytes = (unsigned long)BCMSWAP32(cnt->txbyte);
		s->tx_err_pkts = (uint32_t)BCMSWAP32(cnt->txerror);
		s->tx_dropped_pkts = BCMSWAP32(cnt->txnobuf);

		s->rx_pkts = (unsigned long)BCMSWAP32(cnt->rxframe);
		s->rx_bytes = (unsigned long)BCMSWAP32(cnt->rxbyte);
		s->rx_err_pkts = BCMSWAP32(cnt->rxerror);
		s->rx_dropped_pkts = BCMSWAP32(cnt->rxnobuf);

		s->rx_fcs_err_pkts = BCMSWAP32(cnt->rxcrc);
		s->rx_unknown_pkts = BCMSWAP32(cnt->rxbadproto);
	} else {
		s->tx_pkts = (unsigned long)(cnt->txframe);
		s->tx_bytes = (unsigned long)(cnt->txbyte);
		s->tx_err_pkts = cnt->txerror;
		s->tx_dropped_pkts = cnt->txnobuf;

		s->rx_pkts = (unsigned long)(cnt->rxframe);
		s->rx_bytes = (unsigned long)(cnt->rxbyte);
		s->rx_err_pkts = cnt->rxerror;
		s->rx_dropped_pkts = cnt->rxnobuf;

		s->rx_fcs_err_pkts = cnt->rxcrc;
		s->rx_unknown_pkts = cnt->rxbadproto;
	}

	return 0;
}

int bcmwl_iface_get_stats(const char *ifname, struct wifi_ap_stats *s)
{
	char buf[1024] = {};
	wl_if_stats_t *cnt = (wl_if_stats_t *) buf;
	struct wl_arg arg = {
		.ifname = ifname,
		.cmd = WLC_GET_VAR,
		.iovar = "if_counters",
		.buf = buf,
		.buflen = sizeof(buf),
	};
	int swap;
	int ret;

	libwifi_dbg("[%s] %s called\n", ifname, __func__);
	memset(s, 0, sizeof(*s));

	swap = wl_swap(ifname);
	ret = wl_ctl(&arg);
	if (WARN_ON(ret))
		return ret;

	/* TODO check headers */
	if (swap) {
		s->tx_bytes = (unsigned long)BCMSWAP64(cnt->txbyte);
		s->rx_bytes = (unsigned long)BCMSWAP64(cnt->rxbyte);
		s->tx_pkts = (unsigned long)BCMSWAP64(cnt->txframe);
		s->rx_pkts = (unsigned long)BCMSWAP64(cnt->rxframe);
		s->tx_err_pkts = (uint32_t)BCMSWAP64(cnt->txerror);
		//s->tx_rtx_pkts =
		s->tx_rtx_fail_pkts = (uint32_t)BCMSWAP64(cnt->txfail);
		s->tx_retry_pkts = (uint32_t)BCMSWAP64(cnt->txretry);
		s->tx_mretry_pkts = (uint32_t)BCMSWAP64(cnt->txretrie);
		//s->ack_fail_pkts =
		//s->aggr_pkts =
		s->rx_err_pkts = (uint32_t)BCMSWAP64(cnt->rxerror);
		//s->tx_ucast_pkts =
		//s->rx_ucast_pkts =
		s->tx_dropped_pkts = (uint32_t)BCMSWAP64(cnt->txnobuf);
		s->rx_dropped_pkts = (uint32_t)BCMSWAP64(cnt->rxnobuf);
		s->tx_mcast_pkts = (unsigned long)BCMSWAP64(cnt->txmulti);
		s->rx_mcast_pkts = (unsigned long)BCMSWAP64(cnt->rxmulti);
		//s->tx_bcast_pkts =
		//s->rx_bcast_pkts =
		//s->rx_unknown_pkts =
	} else {
		s->tx_bytes = (unsigned long)cnt->txbyte;
		s->rx_bytes = (unsigned long)cnt->rxbyte;
		s->tx_pkts = (unsigned long)cnt->txframe;
		s->rx_pkts = (unsigned long)cnt->rxframe;
		s->tx_err_pkts = (uint32_t)cnt->txerror;
		//s->tx_rtx_pkts =
		s->tx_rtx_fail_pkts = (uint32_t)cnt->txfail;
		s->tx_retry_pkts = (uint32_t)cnt->txretry;
		s->tx_mretry_pkts = (uint32_t)cnt->txretrie;
		//s->ack_fail_pkts =
		//s->aggr_pkts =
		s->rx_err_pkts = (uint32_t)cnt->rxerror;
		//s->tx_ucast_pkts =
		//s->rx_ucast_pkts =
		s->tx_dropped_pkts = (uint32_t)cnt->txnobuf;
		s->rx_dropped_pkts = (uint32_t)cnt->rxnobuf;
		s->tx_mcast_pkts = (unsigned long)cnt->txmulti;
		s->rx_mcast_pkts = (unsigned long)cnt->rxmulti;
		//s->tx_bcast_pkts =
		//s->rx_bcast_pkts =
		//s->rx_unknown_pkts =
	}

	return ret;
}

int bcmwl_get_supported_security_const(const char *name, uint32_t *sec)
{
	*sec = BIT(WIFI_SECURITY_NONE) |
	       BIT(WIFI_SECURITY_WEP64) |
	       BIT(WIFI_SECURITY_WEP128) |
	       BIT(WIFI_SECURITY_WPA) |
	       BIT(WIFI_SECURITY_WPA2) |
	       BIT(WIFI_SECURITY_WPA3) |
	       BIT(WIFI_SECURITY_WPAPSK) |
	       BIT(WIFI_SECURITY_WPA2PSK) |
	       BIT(WIFI_SECURITY_WPA3PSK);
	return 0;
}

int bcmwl_get_sta_supported_security(const char *name, uint32_t *sec)
{
	/* TODO get this from wpa_supplicant */
	*sec = BIT(WIFI_SECURITY_NONE) |
	       BIT(WIFI_SECURITY_WEP64) |
	       BIT(WIFI_SECURITY_WEP128) |
	       BIT(WIFI_SECURITY_WPA) |
	       BIT(WIFI_SECURITY_WPA2) |
	       BIT(WIFI_SECURITY_WPA3) |
	       BIT(WIFI_SECURITY_WPAPSK) |
	       BIT(WIFI_SECURITY_WPA2PSK) |
	       BIT(WIFI_SECURITY_WPA3PSK);
	return 0;
}

int bcmwl_iface_get_beacon_ies(const char *name, uint8_t *ies, int *len)
{
	uint8_t bcn[WLC_IOCTL_MAXLEN] = { 0 };
	uint8_t *bcn_ies;
	size_t bcn_len = sizeof(bcn);
	size_t ies_len;
	int ret = 0;
	struct beacon_frame *bf;

	libwifi_dbg("[%s] %s called\n", name, __func__);

	if (WARN_ON(!ies) || WARN_ON(!len) || WARN_ON(*len == 0))
		return -EINVAL;

	memset(ies, 0, *len);

	ret = wl_ioctl_iface_get_beacon(name, bcn, &bcn_len);

	if (WARN_ON(ret))
		return ret;

	bf = (struct beacon_frame *)bcn;
	ies_len = bcn_len - ((uint8_t *)&bf->cap_info - bcn + 2);
	bcn_ies = bf->var;

	if (WARN_ON(ies_len <= 0 || *len < ies_len)) {
		*len = 0;
		return -1;
	}

	memcpy(ies, bcn_ies, ies_len);
	*len = (int)ies_len;

	return 0;
}

static int _brcm_get_ap_wmm_params(const char *ifname, uint8_t *p,
				   struct wifi_ap_wmm_ac a[])
{
	a[0].ac = (p[0] & 0x60) >> 5;
	a[0].aifsn = p[0] & 0x0f;
	a[0].cwmin = p[1] & 0x0f;
	a[0].cwmax = (p[1] & 0xf0) >> 4;
	a[0].txop = (p[3] << 8) | p[2];		/* LSB first */

	a[1].ac = (p[4] & 0x60) >> 5;
	a[1].aifsn = p[4] & 0x0f;
	a[1].cwmin = p[5] & 0x0f;
	a[1].cwmax = (p[5] & 0xf0) >> 4;
	a[1].txop = (p[7] << 8) | p[6];

	a[2].ac = (p[8] & 0x60) >> 5;
	a[2].aifsn = p[8] & 0x0f;
	a[2].cwmin = p[9] & 0x0f;
	a[2].cwmax = (p[9] & 0xf0) >> 4;
	a[2].txop = (p[11] << 8) | p[10];

	a[3].ac = (p[12] & 0x60) >> 5;
	a[3].aifsn = p[12] & 0x0f;
	a[3].cwmin = p[13] & 0x0f;
	a[3].cwmax = (p[13] & 0xf0) >> 4;
	a[3].txop = (p[15] << 8) | p[14];

	return 0;
}

int bcmwl_iface_ap_info(const char *name, struct wifi_ap *ap)
{
	int ret = 0;
	uint8_t beacon[WLC_IOCTL_MAXLEN] = { 0 };
	uint8_t *beacon_ies;
	uint8_t *ie_ptr;
	size_t beacon_len = sizeof(beacon);
	size_t ies_len;
	struct beacon_frame *bf;

	/* get beacon */
	ret = wl_ioctl_iface_get_beacon(name, beacon, &beacon_len);

	if (WARN_ON(ret)) {
		return ret;
	}

	bf = (struct beacon_frame *)beacon;
	ies_len = beacon_len - ((uint8_t *)&bf->cap_info - beacon + 2);
	beacon_ies = bf->var;

	ret = wifi_oper_stds_set_from_ie(beacon_ies, ies_len, &ap->bss.oper_std);
	if (ret)
		WARN_ON(bcmwl_get_oper_stds(name, &ap->bss.oper_std));
	WARN_ON(wifi_ssid_advertised_set_from_ie(beacon_ies, ies_len, &ap->ssid_advertised));
	WARN_ON(wifi_get_bss_security_from_ies(&ap->bss, beacon_ies, ies_len));
	WARN_ON(wifi_apload_set_from_ie(beacon_ies, ies_len, &ap->bss.load));
	ap->bss.beacon_int = bf->beacon_int;
	ap->sec.curr_mode = ap->bss.security;

	/* capabilities information */
	ie_ptr = (uint8_t *)&bf->cap_info;
	ap->bss.caps.valid |= WIFI_CAP_BASIC_VALID;
	memcpy(&ap->bss.caps.basic, ie_ptr, sizeof(bf->cap_info));
	wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, sizeof(bf->cap_info));

	/* wmm params */
	ie_ptr = wifi_find_vsie(beacon_ies, ies_len, (uint8_t *)"\x00\x50\xf2", 2, 0xff);
	if (ie_ptr) {
		wifi_cap_set(ap->bss.cbitmap, WIFI_CAP_WMM);
		ie_ptr += 2 + 3 + 2 + 1;  /* skip upto 'version' */
		if (!!(ie_ptr[0] & 0x80))
			wifi_cap_set(ap->bss.cbitmap, WIFI_CAP_APSD);

		ie_ptr += 2; /* points to start of AC parameter records */
		_brcm_get_ap_wmm_params(name, ie_ptr, ap->ac);
	}

	/* ht capabilities */
	ie_ptr = wifi_find_ie(beacon_ies, ies_len, IE_HT_CAP);
	if (ie_ptr) {
		ap->bss.caps.valid |= WIFI_CAP_HT_VALID;
		memcpy(&ap->bss.caps.ht, &ie_ptr[2], ie_ptr[1]);
		wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, ie_ptr[1] + 2);
	}

	/* vht capabilities */
	ie_ptr = wifi_find_ie(beacon_ies, ies_len, IE_VHT_CAP);
	if (ie_ptr) {
		ap->bss.caps.valid |= WIFI_CAP_VHT_VALID;
		memcpy(&ap->bss.caps.vht, &ie_ptr[2], ie_ptr[1]);
		wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, ie_ptr[1] + 2);
	}

	/* extended capabilities */
	ie_ptr = wifi_find_ie(beacon_ies, ies_len, IE_EXT_CAP);
	if (ie_ptr) {
		ap->bss.caps.valid |= WIFI_CAP_EXT_VALID;
		memcpy(&ap->bss.caps.ext, &ie_ptr[2], ie_ptr[1]);
		wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, ie_ptr[1] + 2);
	}

	/* RM enabled capabilities */
	ie_ptr = wifi_find_ie(beacon_ies, ies_len, IE_RRM);
	if (ie_ptr) {
		ap->bss.caps.valid |= WIFI_CAP_RM_VALID;
		memcpy(&ap->bss.caps.ext, &ie_ptr[2], ie_ptr[1]);
		wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, ie_ptr[1] + 2);
	}

	/* mobility domain element */
	ie_ptr = wifi_find_ie(beacon_ies, ies_len, IE_MDE);
	if (ie_ptr) {
		wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, ie_ptr[1] + 2);
	}

	/* HE capabilities */
	ie_ptr = wifi_find_ie_ext(beacon_ies, ies_len, IE_EXT_HE_CAP);
	if (ie_ptr) {
		ap->bss.caps.valid |= WIFI_CAP_HE_VALID;
		memcpy(&ap->bss.caps.he, &ie_ptr[3], min(ie_ptr[1], sizeof(struct wifi_caps_he)));
		wifi_cap_set_from_ie(ap->bss.cbitmap, ie_ptr, ie_ptr[1] + 3);
	}

	return ret;
}

int bcmwl_ap_get_oper_stds(const char *name, uint8_t *std)
{
	uint8_t beacon_ies[WLC_IOCTL_MAXLEN] = { 0 };
	size_t ies_len = sizeof(beacon_ies);
	int ret;

	if (WARN_ON(!std))
		return -EINVAL;

	ret = bcmwl_iface_get_beacon_ies(name, beacon_ies, (int*)&ies_len);

	if (WARN_ON(ret)) {
		return ret;
	}

	ret = wifi_oper_stds_set_from_ie(beacon_ies, ies_len, std);

	if (ret == 0) {
		return 0;
	}

	ret = bcmwl_get_oper_stds(name, std);

	if (WARN_ON(ret)) {
		return ret;
	}

	return 0;
}

int bcmwl_get_oper_band(const char *name, enum wifi_band *band)
{
	uint32_t val, b;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_BAND,
		.buf = &val,
		.buflen = sizeof(val)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);

	ret = wl_ctl(&arg);
	if (ret) {
		*band = BAND_UNKNOWN;
		return ret;
	}

	b = wl_swap_32(name, val);

	switch (b) {
	case 1:
		*band = BAND_5;
		break;
	case 2:
		*band = BAND_2;
		break;
	case 0:
		*band = BAND_DUAL;
		break;
	default:
		*band = BAND_UNKNOWN;
		break;
	}

	return ret;
}

int bcmwl_get_oper_stds(const char *name, uint8_t *std)
{
	enum wifi_band band;
	uint32_t nmode = 0;
	uint32_t vhtmode = 0;
	uint32_t he = 0;

	libwifi_dbg("[%s] %s called\n", name, __func__);

	if (!std)
		return -EINVAL;

	*std = 0;

	WARN_ON(wl_iovar_get(name, "nmode", NULL, 0, &nmode, sizeof(nmode)));
	WARN_ON(wl_iovar_get(name, "vhtmode", NULL, 0, &vhtmode, sizeof(vhtmode)));
	WARN_ON(wl_iovar_get(name, "he", NULL, 0, &he, sizeof(he)));

	if (he) {
		*std |= WIFI_AX;
	}

	if (vhtmode) {
		*std |= WIFI_AC;
	}

	if (nmode) {
		*std |= WIFI_N;
	}

	if (WARN_ON(bcmwl_get_oper_band(name, &band)))
		return 0;

	if (band == BAND_5) {
		*std |= WIFI_A;
	}

	if (band == BAND_2) {
		*std |= WIFI_G;
	}

	return 0;
}

int bcmwl_iface_sta_info(const char *ifname, struct wifi_sta *sta)
{
	libwifi_dbg("[%s] %s called\n", ifname, __func__);

	WARN_ON(wl_cli_get_rssi(ifname, &sta->rssi[0]));
	WARN_ON(wl_cli_get_macaddr(ifname, sta->macaddr));
	WARN_ON(wl_cli_get_bssid(ifname, sta->bssid));
	WARN_ON(bcmwl_get_oper_stds(ifname, &sta->oper_std));

	if (WARN_ON(supplicant_cli_get_sta_supported_security(ifname, &sta->sec.supp_modes)))
		bcmwl_get_sta_supported_security(ifname, &sta->sec.supp_modes);

	WARN_ON(supplicant_cli_get_sta_security(ifname, &sta->sec));

	return 0;
}

int bcmwl_radio_acs(const char *ifname, struct acs_param *p)
{
	char buf[128] = {0};
	int res;

	/* TODO should we start acsd2 if not started or relly on UCI ACS? */
	UNUSED(p);
	res = Cmd(buf, 128, "acs_cli2 -i %s autochannel", ifname);
	WARN_ON(res);

	return res;
}

int bcmwl_radio_channels_info(const char *ifname, struct chan_entry *channel, int *num)
{
	char buf[4096] = {0};
	char *line, *ptr, *nop;
	char cc[3] = {0};
	int nop_minutes;
	int max;
	int chan;
	int res;
	int i;

	max = *num;
	*num = 0;

	memset(channel, 0x0, sizeof(*channel) * max);

	res = Cmd(buf, sizeof(buf), "wl -i %s chan_info", ifname);
	if (WARN_ON(res))
		return -1;

	ptr = buf;
	i = 0;

	bcmwl_radio_get_country(ifname, cc);

	while ((line = strsep(&ptr, "\r\n"))) {
		if (sscanf(line, "Channel %d\n", &chan) != 1)
			continue;

		if (WARN_ON(i >= max))
			return -1;

		channel[i].channel = chan;
		channel[i].freq = c2f(chan);

		do {
			if (!strstr(line, "RADAR Sensitive")) {
				channel[i].dfs = false;
				break;
			}

			channel[i].dfs = true;
			channel[i].dfs_state = WIFI_DFS_STATE_USABLE;
			channel[i].cac_time = cac_time(cc, chan);

			/* Check NOP */
			if ((nop = strstr(line, "Temporarily Out of Service for")) != NULL) {
				if (sscanf(nop, "Temporarily Out of Service for %d minutes", &nop_minutes) == 1)
					channel[i].nop_time = nop_minutes * 60;
				channel[i].dfs_state = WIFI_DFS_STATE_UNAVAILABLE;
				break;
			}

			/* Check ISM */
			if (!strstr(line, "Passive")) {
				/* TODO check if real ISM using dfs_ap_move */
				channel[i].dfs_state = WIFI_DFS_STATE_AVAILABLE;
				break;
			}
		} while (0);

		i++;
	}

	*num = i;
	return 0;
}

int bcmwl_radio_get_beacon_int(const char *name, uint32_t *beacon_int)
{
	int bint;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_BCNPRD,
		.buf = &bint,
		.buflen = sizeof(bint)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);
	*beacon_int = 0;

	ret = wl_vendor(&arg);
	if (ret) {
		libwifi_dbg("[%s] %s fallback to ioctl\n", name, __func__);
		ret = wl_ioctl(&arg);
	}

	if (!ret) {
		*beacon_int = (uint32_t)wl_swap_32(name, bint);
	}

	return ret;
}

int bcmwl_radio_get_dtim(const char *name, uint32_t *dtim_period)
{
	int dtim;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_DTIMPRD,
		.buf = &dtim,
		.buflen = sizeof(dtim)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);
	*dtim_period = 0;

	ret = wl_vendor(&arg);
	if (ret) {
		libwifi_dbg("[%s] %s fallback to ioctl\n", name, __func__);
		ret = wl_ioctl(&arg);
	}

	if (!ret) {
		*dtim_period = (uint32_t)wl_swap_32(name, dtim);
	}

	return ret;
}

int bcmwl_radio_get_txpower(const char *name, int8_t *txpower_dbm, int8_t *txpower_percent)
{
	int val;
	int8_t tmp;

	libwifi_dbg("[%s] %s called\n", name, __func__);

	if (!txpower_dbm || !txpower_percent)
		return -EINVAL;

	*txpower_dbm = 0;
	*txpower_percent = 100;

	//WARN_ON(wl_iovar_get(name, "txpwr1", NULL, 0, &val, sizeof(val)));
	//val &= ~WL_TXPWR_OVERRIDE;
	//temp_val = (char)(val & 0xff);
	//libwifi_dbg("xpumapu txpwr1[%x] temp_val[%x][%d]\n", val, temp_val, temp_val);

	WARN_ON(wl_iovar_get(name, "qtxpower", NULL, 0, &val, sizeof(val)));
	val &= ~WL_TXPWR_OVERRIDE;
	tmp = (int8_t)(val & 0xff);

	*txpower_dbm = tmp/4;
	*txpower_percent = ((128 + tmp) * 100) / 255;

	libwifi_dbg("xpumapu dbm[%d] percent[%d]\n",
		*txpower_dbm,
		*txpower_percent);

	return 0;
}

int bcmwl_radio_get_supp_bw(const char *name, uint32_t *supp_bw)
{
	char buf[WLC_IOCTL_MAXLEN] = {0};
	wl_uint32_list_t *list;
	chanspec_t c;
	int ret;
	int i;
	int swap = wl_swap(name);

	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_VAR,
		.iovar = "chanspecs",
		.buf = &buf[0],
		.buflen = sizeof(buf)
	};

	libwifi_dbg("[%s] %s called\n", name, __func__);
	ret = wl_ctl(&arg);

	if (WARN_ON(ret)) {
		libwifi_dbg("[%s] %s wl_ctl error: ret = %d\n", name, __func__, ret);
		return ret;
	}

	memset(supp_bw, 0, sizeof(uint32_t));
	list = (wl_uint32_list_t *)buf;
	list->count = swap ? BCMSWAP32(list->count) : list->count;
	for (i = 0; i < list->count; i++) {
		c = swap ? (chanspec_t)BCMSWAP32(list->element[i]) :
				(chanspec_t)list->element[i];
		if (CHSPEC_IS20(c)) *supp_bw |= BIT(BW20);
		if (CHSPEC_IS40(c)) *supp_bw |= BIT(BW40);
		if (CHSPEC_IS80(c)) *supp_bw |= BIT(BW80);
		if (CHSPEC_IS160(c)) *supp_bw |= BIT(BW160);
		if (CHSPEC_IS8080(c)) *supp_bw |= BIT(BW8080);
	}
	if (0 == *supp_bw)
		*supp_bw |= BIT(BW_UNKNOWN);

	return ret;
}

/* Events handling */
static const char *event2str(int e)
{
#define C2S(x) case x: return #x
	switch (e) {
	C2S(WLC_E_DEAUTH_IND);
	C2S(WLC_E_DEAUTH);
	C2S(WLC_E_ASSOC);
	C2S(WLC_E_ASSOC_IND);
	C2S(WLC_E_REASSOC);
	C2S(WLC_E_REASSOC_IND);
	C2S(WLC_E_AUTH);
	C2S(WLC_E_AUTH_IND);
	C2S(WLC_E_ACTION_FRAME);
	C2S(WLC_E_ACTION_FRAME_RX);
	C2S(WLC_E_CSA_COMPLETE_IND);
	C2S(WLC_E_AP_CHAN_CHANGE);
	C2S(WLC_E_IF);
	C2S(WLC_E_RADAR_DETECTED);
	C2S(WLC_E_PROBREQ_MSG);
	C2S(WLC_E_RADIO);
	C2S(WLC_E_LINK);
	C2S(WLC_E_PROBREQ_MSG_RX);
	C2S(WLC_E_CAC_STATE_CHANGE);
	}
#undef C2S
	return "";
}

static int bcmwl_event_socket_open(const char *ifname)
{
	struct sockaddr_ll sll;
	int ifindex;
	int fd;

	ifindex = if_nametoindex(ifname);
	if (WARN_ON(ifindex == 0))
		return -1;

	if ((fd = socket(PF_PACKET, SOCK_RAW, htons(ETHER_TYPE_BRCM))) < 0) {
		libwifi_err("[%s] open socket error: %s\n", ifname, strerror(errno));
		return -1;
	}

	memset(&sll, 0, sizeof(sll));
	sll.sll_family   = AF_PACKET;
	sll.sll_protocol = htons(ETHER_TYPE_BRCM);
	sll.sll_ifindex  = ifindex;
	if (bind(fd, (struct sockaddr *) &sll, sizeof(sll)) < 0) {
		libwifi_err("[%s] bind error %s\n", ifname, strerror(errno));
		close(fd);
		return -1;
	}

	return fd;
}

static int bcmwl_event_socket_close(int fd)
{
    return close(fd);
}

static ssize_t bcmwl_event_socket_read(int fd, void *msg, size_t msglen)
{
	return recv(fd, msg, msglen, MSG_DONTWAIT);
}

/* Internal bcmwl event context */
struct bcmwl_event_ctx {
	struct event_struct ev;
	int fd;
	int resp_max_len;
};

int bcmwl_register_event(const char *ifname, struct event_struct *ev,
			  void **evhandle)
{
	struct bcmwl_event_ctx *ctx;
	int ret = -1;

	libwifi_dbg("[%s] %s called\n", ifname, __func__);

	*evhandle = NULL;

	ctx = calloc(1, sizeof(*ctx));
	if (WARN_ON(!ctx))
		return ret;

	ev->fd_monitor = bcmwl_event_socket_open(ifname);
	if (ev->fd_monitor == -1) {
		libwifi_err("[%s] %s open socket failed %d\n", ifname, __func__, errno);
		free(ctx);
		return ret;
	}

	memcpy(&ctx->ev, ev, sizeof(*ev));
	ctx->fd = ev->fd_monitor;
	ctx->resp_max_len = ev->resp.len;

	*evhandle = ctx;

	/* Register events we would like report to upper layer */
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_DEAUTH_IND));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_DEAUTH));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_ASSOC_IND));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_ASSOC));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_REASSOC_IND));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_REASSOC));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_AUTH_IND));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_AUTH));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_ACTION_FRAME));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_ACTION_FRAME_RX));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_CSA_COMPLETE_IND));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_AP_CHAN_CHANGE));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_RADAR_DETECTED));
	WARN_ON(bcmwl_enable_event_bit(ifname, WLC_E_CAC_STATE_CHANGE));

	return 0;
}

int bcmwl_unregister_event(const char *ifname, void *evhandle)
{
	struct bcmwl_event_ctx *ctx = evhandle;

	libwifi_dbg("[%s] %s called\n", ifname, __func__);
	bcmwl_event_socket_close(ctx->fd);
	free(ctx);

	/*
	 * Don't disable event bits, while we could have some
	 * other user for that. Found that disable all events
	 * introduce issue(s) for DHD implementation, next
	 * hostapd won't get cfg80211 events.
	 */

	return 0;
}

static int bcmwl_event_handle(const char *ifname, bcm_event_t *event, int size, char *buf, size_t buf_size)
{
	char macaddr[18] = {0};
	char raw[1024] = {0};
	char event_name[128] = {0};
	uint32_t event_type;
	uint32_t datalen;

	datalen = ntohl(event->event.datalen);
	event_type = ntohl(event->event.event_type);

	switch (event_type) {
		case WLC_E_DEAUTH_IND:
		case WLC_E_DEAUTH:
			snprintf(event_name, sizeof(event_name), "deauth");
			break;
		case WLC_E_ASSOC:
		case WLC_E_ASSOC_IND:
			snprintf(event_name, sizeof(event_name), "assoc");
			break;
		case WLC_E_REASSOC:
		case WLC_E_REASSOC_IND:
			snprintf(event_name, sizeof(event_name), "reassoc");
			break;
		case WLC_E_AUTH:
		case WLC_E_AUTH_IND:
			snprintf(event_name, sizeof(event_name), "auth");
			break;
		case WLC_E_ACTION_FRAME:
		case WLC_E_ACTION_FRAME_RX:
			snprintf(event_name, sizeof(event_name), "action");
			break;
		/* we already have this from hostapd/supplicant */
		case WLC_E_AUTHORIZED:
		default:
			return -1;
	}

	libwifi_dbg("[%s] %s called event %u (%s) iface %s size %d datalen %u\n",
		    ifname, __func__, event_type, event_name, event->event.ifname,
		    size, datalen);

	/* Check frame length */
	if (WARN_ON(sizeof(*event) + datalen >= size))
		return 0;

	/* Check if we have place for raw data */
	if (datalen * 2 < sizeof(raw))
		btostr((uint8_t *) (event + 1), datalen, raw);

	/* Fill macaddr */
	snprintf(macaddr, sizeof(macaddr), MACSTR, MAC2STR(event->event.addr.octet));

	/* TODO agree format with upper layer */
	snprintf(buf, buf_size,
		 "wifi.sta '{\"ifname\":\"%s\", \"event\":\"%s\", \"data\":{\"macaddr\":\"%s\", \"raw\":\"%s\"}}'",
		 event->event.ifname, event_name, macaddr, raw);

	return 0;
}

static void bcmwl_get_chanspec(chanspec_t chanspec, char *chan, size_t chan_size, char *bw, size_t bw_size)
{
	snprintf(chan, chan_size, "%u", chanspec_to_ctrlchannel(chanspec));
	if (CHSPEC_IS20(chanspec))
		snprintf(bw, bw_size, "20");
	if (CHSPEC_IS40(chanspec))
		snprintf(bw, bw_size, "40");
	if (CHSPEC_IS80(chanspec))
		snprintf(bw, bw_size, "80");
	if (CHSPEC_IS160(chanspec))
		snprintf(bw, bw_size, "160");
	if (CHSPEC_IS8080(chanspec))
		snprintf(bw, bw_size, "8080");
}

static int bcmwl_event_handle_csa(const char *ifname, bcm_event_t *event, int size, char *buf, size_t buf_size)
{
	/* TODO Seems today driver don't send target chanspec, fix it */
	snprintf(buf, buf_size,
		 "wifi.sta '{\"ifname\":\"%s\", \"event\":\"%s\", \"data\": {\"status\": \"%d\"}}'",
		 event->event.ifname, "ap-csa-finished", ntohl(event->event.reason));

	return 0;
}

static int bcmwl_event_handle_radar(const char *ifname, bcm_event_t *event, int size, char *buf, size_t buf_size)
{
	wl_event_radar_detect_data_t *ev;
	char chan[32], tchan[32];
	char bw[32], tbw[32];
	int swap;

	ev = (wl_event_radar_detect_data_t *) (event + 1);

	swap = wl_swap(ifname);

	if (swap) {
		ev->current_chanspec = BCMSWAP16(ev->current_chanspec);
		ev->target_chanspec = BCMSWAP16(ev->target_chanspec);
	}

	bcmwl_get_chanspec(ev->current_chanspec, chan, sizeof(chan), bw, sizeof(bw));
	bcmwl_get_chanspec(ev->target_chanspec, tchan, sizeof(tchan), tbw, sizeof(tbw));

	snprintf(buf, buf_size,
		 "wifi.sta '{\"ifname\":\"%s\", \"event\":\"%s\", \"data\": {\"channel\":\"%s\", \"width\":\"%s\", \"target-channel\":\"%s\", \"target-width\":\"%s\"}}'",
		 event->event.ifname, "radar", chan, bw, tchan, tbw);

	return 0;
}

static int bcmwl_event_handle_chan_change(const char *ifname, bcm_event_t *event, int size, char *buf, size_t buf_size)
{
	wl_event_change_chan_t *ev;
	char event_name[64] = {0};
	char reason[64] = {0};
	unsigned int length;
	char chan[32];
	char bw[32] = "20";
	int swap;

	ev = (wl_event_change_chan_t *)(event + 1);
	length = ntohl(event->event.datalen);

	if (WARN_ON(length < sizeof(*ev)))
		return -1;

	/* Header - bcm_event_t using network endianness */
	swap = wl_swap(ifname);

	snprintf(event_name, sizeof(event_name), "ap-chan-change");

	/* Data after bcm_event_t using HW endianness */
	if (swap) {
		ev->reason = BCMSWAP32(ev->reason);
		ev->target_chanspec = BCMSWAP16(ev->target_chanspec);
	}

	switch (ev->reason) {
	case WL_CHAN_REASON_CSA:
		snprintf(reason, sizeof(reason), "csa");
		break;
	case WL_CHAN_REASON_DFS_AP_MOVE_START:
		snprintf(reason, sizeof(reason), "move-start");
		break;
	case WL_CHAN_REASON_DFS_AP_MOVE_RADAR_FOUND:
		snprintf(reason, sizeof(reason), "radar");
		break;
	case WL_CHAN_REASON_DFS_AP_MOVE_ABORTED:
		snprintf(reason, sizeof(reason), "move-aborted");
		break;
	case WL_CHAN_REASON_DFS_AP_MOVE_SUCCESS:
		snprintf(reason, sizeof(reason), "move-success");
		break;
	case WL_CHAN_REASON_DFS_AP_MOVE_STUNT:
		snprintf(reason, sizeof(reason), "move-stunt");
		break;
	case WL_CHAN_REASON_CSA_TO_DFS_CHAN_FOR_CAC_ONLY:
		snprintf(reason, sizeof(reason), "cac-only");
		break;
	case WL_CHAN_REASON_ANY:
		snprintf(reason, sizeof(reason), "any");
		break;
	default:
		libwifi_err("[%s] %s unhandled reason %d\n", ifname, __func__, ev->reason);
		return -1;
	}

	bcmwl_get_chanspec(ev->target_chanspec, chan, sizeof(chan), bw, sizeof(bw));

	snprintf(buf, buf_size,
		 "wifi.ap '{\"ifname\":\"%s\", \"event\":\"%s\", \"data\": {\"reason\":\"%s\", \"target-channel\":\"%s\", \"target-width\":\"%s\"}}'",
		 event->event.ifname, event_name, reason, chan, bw);

	return 0;
}

static int bcmwl_event_handle_if(const char *ifname, bcm_event_t *event, int size, char *buf, size_t buf_size)
{
	struct wl_event_data_if *ev;
	char event_name[64] = {0};
	char macaddr[18] = {0};
	char role[64] = {0};
	unsigned int length;
	char name[16] = {0};

	ev = (struct wl_event_data_if *)(event + 1);
	length = ntohl(event->event.datalen);

	if (WARN_ON(length < sizeof(*ev)))
		return -1;

	switch (ev->opcode) {
	case WLC_E_IF_ADD:
		snprintf(event_name, sizeof(event_name), "iface-add");
		break;
	case WLC_E_IF_DEL:
		snprintf(event_name, sizeof(event_name), "iface-del");
		break;
	case WLC_E_IF_CHANGE:
		snprintf(event_name, sizeof(event_name), "iface-change");
		break;
	case WLC_E_IF_BSSCFG_UP:
		snprintf(event_name, sizeof(event_name), "bss-up");
		break;
	case WLC_E_IF_BSSCFG_DOWN:
		snprintf(event_name, sizeof(event_name), "bss-down");
		break;
	default:
		return -1;
	}

	switch (ev->role) {
	case WLC_E_IF_ROLE_STA:
		snprintf(role, sizeof(role), "sta");
		break;
	case WLC_E_IF_ROLE_AP:
		snprintf(role, sizeof(role), "ap");
		break;
	case WLC_E_IF_ROLE_WDS:
		snprintf(role, sizeof(role), "apvlan");
		break;
	default:
		break;
	}

	snprintf(macaddr, sizeof(macaddr), MACSTR, MAC2STR(ev->peer_addr.octet));

	/* TODO check if this is correct */
	snprintf(name, sizeof(name), "wds%u.0.%u", ev->bssidx, ev->ifidx);

	libwifi_dbg("[%s] %s opcode %s role %s ifidx %u bssidx %u macaddr %s [%s]\n",
		    ifname, __func__, event_name, role, ev->ifidx, ev->bssidx, macaddr, name);

	snprintf(buf, buf_size,
		 "wifi.ap '{\"ifname\":\"%s\", \"event\":\"%s\", \"data\": {\"role\":\"%s\", \"macaddr\":\"%s\"}}'",
		 event->event.ifname, event_name, role, macaddr);

	/* Send to upper layer - seems today only DHD report this event
	 * TODO fix it
	 */
	switch (ev->role) {
	case WLC_E_IF_ROLE_WDS:
		switch (ev->opcode) {
		case WLC_E_IF_ADD:
			snprintf(buf, buf_size,
				 "wifi.sta '{\"ifname\":\"%s\", \"event\": \"wds-station-added\", \"data\" : {\"ifname\":\"%s\", \"macaddr\":\"%s\"}}'",
				 ifname, name, macaddr);
			break;
		case WLC_E_IF_DEL:
			snprintf(buf, buf_size,
				 "wifi.sta '{\"ifname\":\"%s\", \"event\": \"wds-station-removed\", \"data\" : {\"ifname\":\"%s\", \"macaddr\":\"%s\"}}'",
				 ifname, name, macaddr);
			break;
		default:
			break;
		}
	default:
		break;
	}

        return 0;
}


static const char *dfs_state2str(int e)
{
#define C2S(x) case x: return #x
	switch (e) {
	C2S(WL_DFS_CACSTATE_IDLE);
	C2S(WL_DFS_CACSTATE_PREISM_CAC);
	C2S(WL_DFS_CACSTATE_ISM);
	C2S(WL_DFS_CACSTATE_CSA);
	C2S(WL_DFS_CACSTATE_POSTISM_CAC);
	C2S(WL_DFS_CACSTATE_PREISM_OOC);
	C2S(WL_DFS_CACSTATE_POSTISM_OOC);
	}
#undef C2S
	return "unknown";
}

static int bcmwl_event_handle_cac_state_change(const char *ifname, bcm_event_t *event, int size, struct bcmwl_event_ctx *ctx)
{
	wlc_cac_event_t *ev;
	wl_dfs_status_all_t *scan_status;
	wl_dfs_sub_status_t *status;
	char out[2048] = {0};
	char event_name[64] = {0};
	unsigned int length;
	struct event_response *resp;
	char chan[64] = {0};
	char bw[64] = {0};
	int resp_max;
	int out_len;
	int swap;
	int i;

	ev = (wlc_cac_event_t *)(event + 1);
	length = ntohl(event->event.datalen);

	if (WARN_ON(length < sizeof(*ev)))
		return -1;

	resp = &ctx->ev.resp;
	resp_max = ctx->resp_max_len;

	swap = wl_swap(ifname);

	scan_status = &ev->scan_status;
	if (swap) {
		ev->version = BCMSWAP16(ev->version);
		ev->length = BCMSWAP16(ev->length);
		ev->type = BCMSWAP16(ev->type);

		/* TODO check other versions */
		if (WARN_ON(ev->version != WLC_E_CAC_STATE_TYPE_DFS_STATUS_ALL))
			return -1;

		scan_status->version = BCMSWAP16(scan_status->version);
		scan_status->num_sub_status = BCMSWAP16(scan_status->num_sub_status);

		for (i = 0; i < scan_status->num_sub_status; i++) {
			status = &scan_status->dfs_sub_status[i];
			status->state = BCMSWAP32(status->state);
			status->duration = BCMSWAP32(status->duration);
			status->chanspec = BCMSWAP16(status->chanspec);
			status->chanspec_last_cleared = BCMSWAP16(status->chanspec_last_cleared);
			status->sub_type = BCMSWAP16(status->sub_type);
		}
	}

	for (i = 0; i < scan_status->num_sub_status; i++) {
		status = &scan_status->dfs_sub_status[i];
		bcmwl_get_chanspec(status->chanspec, chan, sizeof(chan), bw, sizeof(bw));
		libwifi_dbg("[%s] [%d] chan %s bw %s state %u %s\n", ifname, i, chan, bw, status->state, dfs_state2str(status->state));

		snprintf(out, sizeof(out),
			 "wifi.sta '{\"ifname\":\"%s\", \"event\": \"%s\", \"data\" : {\"channel\":\"%s\", \"bw\":\"%s\"}}'",
			 event->event.ifname, dfs_state2str(status->state), chan, bw);

		out_len = strlen(out);
		if (WARN_ON(out_len >= resp_max))
			return 0;

		resp->type = WIFI_EVENT_VENDOR;
		memcpy(resp->data, out, out_len);
		resp->len = out_len;

		if (ctx->ev.cb)
			ctx->ev.cb(&ctx->ev);
	}

	return 0;
}

int bcmwl_recv_event(const char *ifname, void *evhandle)
{
	struct bcmwl_event_ctx *ctx = evhandle;
	char buf[2048] = {0};
	uint32_t event_type;
	bcm_event_t *event;
	int frame_size;

	struct event_response *resp;
	char out[2048] = {0};
	int resp_max;
	int out_len;
	int res;

	if (WARN_ON(!ctx))
		return -1;

	frame_size = bcmwl_event_socket_read(ctx->fd, buf, sizeof(buf));
	if (frame_size < 0) {
		if (errno != EAGAIN)
			libwifi_dbg("[%s] %s read error %s (%d)\n", ifname, __func__, strerror(errno), errno);
		return -errno;
	}

	if (WARN_ON(frame_size < sizeof(bcm_event_t)))
		return -1;

	/* This is what we get from lower layer */
	event = (bcm_event_t *) buf;
	event_type = ntohl(event->event.event_type);

	/* To upper layer */
	resp = &ctx->ev.resp;
	resp_max = ctx->resp_max_len;

	if (event_type != WLC_E_PROBREQ_MSG_RX &&
	    event_type != WLC_E_PROBREQ_MSG)
		libwifi_dbg("[%s] event %d (%s) size %d\n", ifname, event_type, event2str(event_type), frame_size);

	switch (event_type) {
	case WLC_E_DEAUTH_IND:
	case WLC_E_DEAUTH:
	case WLC_E_ASSOC:
	case WLC_E_ASSOC_IND:
	case WLC_E_REASSOC:
	case WLC_E_REASSOC_IND:
	case WLC_E_AUTH:
	case WLC_E_AUTH_IND:
	case WLC_E_ACTION_FRAME:
	case WLC_E_ACTION_FRAME_RX:
		res = bcmwl_event_handle(ifname, event, frame_size, out, sizeof(out));
		break;
	case WLC_E_CSA_COMPLETE_IND:
		res = bcmwl_event_handle_csa(ifname, event, frame_size, out, sizeof(out));
		break;
	case WLC_E_AP_CHAN_CHANGE:
		res = bcmwl_event_handle_chan_change(ifname, event, frame_size, out, sizeof(out));
		break;
	case WLC_E_IF:
		res = bcmwl_event_handle_if(ifname, event, frame_size, out, sizeof(out));
		break;
	case WLC_E_RADAR_DETECTED:
		res = bcmwl_event_handle_radar(ifname, event, frame_size, out, sizeof(out));
		break;
	case WLC_E_CAC_STATE_CHANGE:
		/* Handle this internally, while we will send two events */
		return bcmwl_event_handle_cac_state_change(ifname, event, frame_size, ctx);
	case WLC_E_PROBREQ_MSG:
	case WLC_E_RADIO:
	case WLC_E_LINK:
	case WLC_E_PROBREQ_MSG_RX:
	default:
		res = -1;
		break;
	}

	if (res)
		return res;

	/* Pass to upper layer */
	out_len = strlen(out);
	if (WARN_ON(out_len >= resp_max))
		return 0;

	resp->type = WIFI_EVENT_VENDOR;
	memcpy(resp->data, out, out_len);
	resp->len = out_len;

	if (ctx->ev.cb)
		ctx->ev.cb(&ctx->ev);

	return 0;
}

static int bcmwl_event_bit(const char *ifname, unsigned int bit, int set)
{
	int mask_len = (WLC_E_LAST + 7) / 8; // WL_EVENTING_MASK_EXT_LEN
	eventmsgs_ext_t *ext;
	int ret = -1;

	ext = calloc(1, sizeof(*ext) + mask_len);
	if (WARN_ON(!ext))
		return -1;

	if (WARN_ON((bit / 8) >= mask_len))
		return -1;

	ext->ver = EVENTMSGS_VER;
	ext->len = mask_len;

	if (set) {
		ext->mask[bit/8] = (1 << (bit % 8));
		ext->command = EVENTMSGS_SET_BIT;
	} else {
		memset(ext->mask, 0xff, mask_len);
		ext->mask[bit/8] = ~(1 << (bit % 8));
		ext->command = EVENTMSGS_RESET_BIT;
	}

	ret = wl_iovar_set(ifname, "event_msgs_ext", ext, sizeof(*ext) + mask_len, NULL, 0);

	free(ext);
	return ret;
}

int bcmwl_enable_event_bit(const char *ifname, unsigned int bit)
{
	return bcmwl_event_bit(ifname, bit, 1);
}

int bcmwl_disable_event_bit(const char *ifname, unsigned int bit)
{
	return bcmwl_event_bit(ifname, bit, 0);
}

int bcmwl_driver_info(const char *name, struct wifi_metainfo *info)
{
	size_t len = (sizeof(((struct wifi_metainfo*)0)->fw_data));

	libwifi_dbg("[%s] %s called\n", name, __func__);

	nlwifi_driver_info(name, info);
	wl_ioctl_get_version(name, info->fw_data, &len);

	return 0;
}

int bcmwl_iface_get_wpa_auth(const char *name, uint32_t *wpa_auth)
{
	uint32_t val;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_WPA_AUTH,
		.buf = &val,
		.buflen = sizeof(val)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);

	ret = wl_ctl(&arg);
	if (!ret) {
		*wpa_auth = wl_swap_32(name, val);
		return ret;
	}

	return ret;
}

int bcmwl_iface_get_auth(const char *name, uint32_t *auth)
{
	uint32_t val;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_AUTH,
		.buf = &val,
		.buflen = sizeof(val)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);

	ret = wl_ctl(&arg);
	if (!ret) {
		*auth = wl_swap_32(name, val);
		return ret;
	}

	return ret;
}

int bcmwl_iface_get_wsec(const char *name, uint32_t *enc)
{
	uint32_t val;
	int ret;
	struct wl_arg arg = {
		.ifname = name,
		.cmd = WLC_GET_WSEC,
		.buf = &val,
		.buflen = sizeof(val)
	};

	libwifi_dbg("[%s] %s called (vendor)\n", name, __func__);

	ret = wl_ctl(&arg);
	if (!ret) {
		*enc = wl_swap_32(name, val);
		return ret;
	}

	return ret;
}


static int bcm_get_sta_info_airtime(const char *ifname, uint8_t *addr,
							struct wifi_sta *wsta)
{
	char bufptr[1536] = {0};
	iov_bs_data_struct_t *data = (iov_bs_data_struct_t *)bufptr;
	int32_t flag = 0;
	int num_stas = 0;
	int swap;
	int i;

	if (!wsta)
		return -1;

	swap = wl_swap(ifname);

	if (wl_iovar_get(ifname, "bs_data", &flag, sizeof(flag), bufptr, 1536) < 0) {
		libwifi_err("wl_iovar_get error!!\n");
		return -1;
	}

	num_stas = swap ? BCMSWAP16(data->structure_count) : data->structure_count;
	for (i = 0; i < num_stas; i++) {
		iov_bs_data_record_t *rec = &data->structure_record[i];
		iov_bs_data_counters_t *ctr = &rec->station_counters;

		if (!memcmp(rec->station_address.octet, addr, 6)) {
			int airtm = 0;

			ctr->time_delta = swap ? BCMSWAP32(ctr->time_delta) :
					ctr->time_delta;

			ctr->airtime = swap ? BCMSWAP32(ctr->airtime) :
					ctr->airtime;

			airtm = ctr->time_delta ?
				ctr->airtime * 100 / ctr->time_delta : 0;

			wsta->airtime = airtm;
			if (wsta->airtime > 100) {
				/* Rare but seen; can happen when more than
				 * 1's worth of sampled data is considered.
				 */
				wsta->airtime = 100;
			}

			return 0;
		}
	}
	return 0;
}

int bcmwl_iface_get_assoclist(const char *ifname, uint8_t *stas, int *num_stas)
{
	char buf[WLC_IOCTL_MAXLEN] = {0};
	struct maclist *maclist;
	struct wl_arg arg = {
		.ifname = ifname,
		.cmd = WLC_GET_ASSOCLIST,
		.buf = &buf,
		.buflen = sizeof(buf)
	};
	int swap;
	int max;
	int i;

	max = (WLC_IOCTL_MAXLEN - sizeof(int)) / ETHER_ADDR_LEN;
	maclist = (struct maclist *) buf;
	swap = wl_swap(ifname);

	maclist->count = swap ? BCMSWAP32(max) : max;
	if (WARN_ON(wl_ioctl(&arg)))
		return -1;

	maclist->count = swap ? BCMSWAP32(maclist->count) : maclist->count;
	for (i = 0; i < maclist->count; i++) {
		if (WARN_ON(i >= *num_stas))
			return -1;

		memcpy(&stas[i * 6], maclist->ea[i].octet, 6);
	}

	*num_stas = i;
	return 0;
}

static char * he_get_table(int index)
{
	switch (index) {
	case 0:
		return "bw80_tx";
	case 1:
		return "bw80_rx";
	case 2:
		return "bw160_tx";
	case 3:
		return "bw160_rx";
	case 4:
		return "bw80p80_tx";
	case 5:
		return "bw80p80_rx";
	default:
		return "unknown";
	}
}

static char * he_get_mcs(uint8_t bits)
{
	switch (bits) {
	case 0:
		return "mcs0_7";
	case 1:
		return "mcs0_9";
	case 2:
		return "mcs0_11";
	default:
		return "unknown";
	}
}

int bcmwl_get_sta_info(const char *ifname, uint8_t *addr, struct wifi_sta *wsta)
{
	sta_info_v8_t *sta;
	char buf[WLC_IOCTL_MAXLEN] = {0};
	int bw = 20;
	int sgi = 0;
	int nss = 0;
	int max_mcs = -1;
	int swap;
	int i;
	int j;


	if (wsta == NULL)
		return -1;

	swap = wl_swap(ifname);

	if (wl_iovar_get(ifname, "sta_info", addr, 6, buf, sizeof(buf)) < 0) {
		libwifi_err("wl_iovar_get error!!\n");
		return -1;
	}

	sta = (sta_info_v8_t *) buf;
	if (swap)
		sta->ver = BCMSWAP16(sta->ver);

	if (WARN_ON(sta->ver != WL_STA_VER_8))
		return -1;

	if (swap) {
		sta->len = BCMSWAP16(sta->len);
		sta->cap = BCMSWAP16(sta->cap);
		sta->aid = BCMSWAP16(sta->aid);
		sta->flags = BCMSWAP32(sta->flags);
		sta->idle = BCMSWAP32(sta->idle);
		sta->rateset.count = BCMSWAP32(sta->rateset.count);
		sta->in = BCMSWAP32(sta->in);
		sta->listen_interval_inms =
				BCMSWAP32(sta->listen_interval_inms);
		sta->ht_capabilities = BCMSWAP16(sta->ht_capabilities);
		sta->vht_flags = BCMSWAP16(sta->vht_flags);
		sta->he_flags = BCMSWAP32(sta->he_flags);
		sta->he_omi = BCMSWAP16(sta->he_omi);

		if (sta->flags & WL_STA_SCBSTATS) {
			sta->tx_tot_pkts = BCMSWAP32(sta->tx_tot_pkts);
			sta->tx_tot_bytes = BCMSWAP64(sta->tx_tot_bytes);
			sta->tx_pkts = BCMSWAP32(sta->tx_pkts);
			sta->tx_ucast_bytes = BCMSWAP64(sta->tx_ucast_bytes);
			sta->tx_mcast_pkts = BCMSWAP32(sta->tx_mcast_pkts);
			sta->tx_mcast_bytes = BCMSWAP64(sta->tx_mcast_bytes);
			sta->tx_failures = BCMSWAP32(sta->tx_failures);
			sta->rx_tot_pkts = BCMSWAP32(sta->rx_tot_pkts);
			sta->rx_tot_bytes = BCMSWAP64(sta->rx_tot_bytes);
			sta->rx_ucast_pkts = BCMSWAP32(sta->rx_ucast_pkts);
			sta->rx_ucast_bytes = BCMSWAP64(sta->rx_ucast_bytes);
			sta->rx_mcast_pkts = BCMSWAP32(sta->rx_mcast_pkts);
			sta->rx_mcast_bytes = BCMSWAP64(sta->rx_mcast_bytes);
			sta->tx_rate = BCMSWAP32(sta->tx_rate);
			sta->tx_rate_fallback =
				BCMSWAP32(sta->tx_rate_fallback);
			sta->rx_rate = BCMSWAP32(sta->rx_rate);
			sta->rx_decrypt_succeeds =
				BCMSWAP32(sta->rx_decrypt_succeeds);
			sta->rx_decrypt_failures =
				BCMSWAP32(sta->rx_decrypt_failures);
			sta->tx_pkts_retried = BCMSWAP32(sta->tx_pkts_retried);
			sta->tx_pkts_total = BCMSWAP32(sta->tx_pkts_total);
			sta->tx_pkts_retries = BCMSWAP32(sta->tx_pkts_retries);
			sta->tx_pkts_retry_exhausted =
				BCMSWAP32(sta->tx_pkts_retry_exhausted);
			sta->tx_pkts_fw_total =
				BCMSWAP32(sta->tx_pkts_fw_total);
			sta->tx_pkts_fw_retries =
				BCMSWAP32(sta->tx_pkts_fw_retries);
			sta->tx_pkts_fw_retry_exhausted =
				BCMSWAP32(sta->tx_pkts_fw_retry_exhausted);
			sta->rx_pkts_retried = BCMSWAP32(sta->rx_pkts_retried);
		}
	}

	/* fill wifi_sta */
	memset(wsta, 0, sizeof(struct wifi_sta));
	memcpy(wsta->macaddr, sta->ea.octet, 6);

	/* libwifi_dbg("Capability: 0x%04x\n", sta->cap); */
	wifi_cap_set_from_ie(wsta->cbitmap, (uint8_t *)&sta->cap, 2); /* set caps bits */

	if (!!(sta->flags & WL_STA_N_CAP)) {
		wsta->caps.valid |= WIFI_CAP_HT_VALID;

		if (!!(sta->ht_capabilities & WL_STA_CAP_LDPC_CODING))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HT_LDPC);

		if (!!(sta->ht_capabilities & WL_STA_CAP_40MHZ))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_2040);

		if (!!(sta->ht_capabilities & WL_STA_CAP_GF))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HT_GREENFIELD);

		if (!!(sta->ht_capabilities & WL_STA_CAP_SHORT_GI_20)) {
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_SGI20);
			sgi = 1;
		}

		if (!!(sta->ht_capabilities & WL_STA_CAP_SHORT_GI_40)) {
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_SGI40);
			sgi = 1;
		}

		if (!!(sta->ht_capabilities & WL_STA_CAP_TX_STBC))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HT_TX_STBC);

		if (!!(sta->ht_capabilities & WL_STA_CAP_RX_STBC_MASK))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HT_RX_STBC);

		if (!!(sta->ht_capabilities & WL_STA_CAP_MIMO_PS_MASK))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HT_SMPS);

		/* set wmm cap true for ht sta */
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_WMM);

		{
			uint8_t *supp_mcs = sta->rateset_adv.mcs;
			int octet = 0;
			int l;

			for (l = 0; l < 76; l++) {
				if (l && !(l % 8))
					octet++;

				if (!!(supp_mcs[octet] & bit(l % 8)))
					max_mcs++;
			}
			nss = (max_mcs / 8) + 1;
			max_mcs %= 8;

			bw = 20;
			if (sta->link_bw > 0 && sta->link_bw <= 4)
				bw *= (1 << (sta->link_bw - 1));

			if (max_mcs > 0) {
				wsta->rate.rate = wifi_mcs2rate(max_mcs, bw, nss, sgi);
				wsta->rate.m.mcs = max_mcs;
				wsta->maxrate = wsta->rate.rate;
			}
		}
	}

	if (!!(sta->flags & WL_STA_VHT_CAP)) {
		wsta->caps.valid |= WIFI_CAP_VHT_VALID;

		if (!!(sta->vht_flags & WL_STA_VHT_LDPCCAP))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_VHT_RX_LDPC);

		if (!!(sta->vht_flags & WL_STA_SGI80)) {
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_SGI80);
			sgi = 1;
		}

		if (!!(sta->vht_flags & WL_STA_SGI160)) {
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_SGI160);
			sgi = 1;
		}

		if (!!(sta->vht_flags & WL_STA_VHT_TX_STBCCAP))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_VHT_TX_STBC);

		if (!!(sta->vht_flags & WL_STA_SU_BEAMFORMEE))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_VHT_SU_BFE);

		if (!!(sta->vht_flags & WL_STA_SU_BEAMFORMER))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_VHT_SU_BFR);

		if (!!(sta->vht_flags & WL_STA_MU_BEAMFORMEE))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_VHT_MU_BFE);

		if (!!(sta->vht_flags & WL_STA_MU_BEAMFORMER))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_VHT_MU_BFR);

		/* set wmm cap true for vht */
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_WMM);

		{
			nss = 0;
			max_mcs = 0;
			for (i = 0; i < WL_VHT_CAP_MCS_MAP_NSS_MAX; i++) {
				uint16_t vht_mcs;

				vht_mcs = swap ?
					BCMSWAP16(sta->rateset_adv.vht_mcs[i])
					: sta->rateset_adv.vht_mcs[i];

				if (!vht_mcs)
					break;

				if ((vht_mcs & 0x3ff) == 0x3ff)
					max_mcs = 9;
				else if ((vht_mcs & 0x3ff) == 0x1ff)
					max_mcs = 8;
				else
					max_mcs = 7;

				nss++;
			}

			bw = 20;
			if (sta->link_bw > 0 && sta->link_bw <= 4)
				bw *= (1 << (sta->link_bw - 1));

			libwifi_dbg("max_mcs %d bw %d nss %d sgi %d\n", max_mcs, bw, nss, sgi);
			if (max_mcs) {
				wsta->rate.rate = wifi_mcs2rate(max_mcs, bw, nss, sgi);
				wsta->rate.m.mcs = max_mcs;
				wsta->maxrate = wsta->rate.rate;
			}
		}
	}

	if (!!(sta->flags & WL_STA_HE_CAP)) {
		wsta->caps.valid |= WIFI_CAP_HE_VALID;

		libwifi_dbg("he_flags: 0x%x\n", sta->he_flags);

		/* TODO check more flags here */
		if (!!(sta->he_flags & WL_STA_HE_LDPCCAP))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_LDPC_PAYLOAD);

		if (!!(sta->he_flags & WL_STA_HE_TX_STBCCAP))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_STBC_TX_80);

		if (!!(sta->he_flags & WL_STA_HE_RX_STBCCAP))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_STBC_RX_80);

		//if (!!(sta->he_flags & WL_STA_HE_HTC_CAP))
		//	wifi_cap_set(wsta->cbitmap, );

		if (!!(sta->he_flags & WL_STA_HE_SU_MU_BEAMFORMEE)) {
		//	wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_MU_BFE);
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_SU_BFE);
		}

		if (!!(sta->he_flags & WL_STA_HE_MU_BEAMFORMER))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_MU_BFR);

		if (!!(sta->he_flags & WL_STA_HE_SU_BEAMFORMER))
			wifi_cap_set(wsta->cbitmap, WIFI_CAP_HE_SU_BFR);

		for (i = 0; i < 6; i++) {
			uint16_t he_mcs;
			uint8_t bits;

			/*
			 * Index in table:
			 *  [0] = bw80_tx
			 *  [1] = bw80 rx
			 *  [2] = bw160 tx
			 *  [3] = bw160 rx
			 *  [4] = bw80p80 tx
			 *  [5] = bw80p80 rx
			 *
                         * he_mcs format:
			 *  Two bits per stream: d[1:0] is for stream #0, d[3:2] for stream #1, etc.
			 *  'b00: mandatory minimum support of MCS 0 through 7
			 *  'b01: support of MCS 0 through 8
			 *  'b10: support of MCS 0 through 9
			 *  'b11: no HE MCS'es supported
			 */

			he_mcs = swap ?
				BCMSWAP16(sta->rateset_adv.he_mcs[i])
				: sta->rateset_adv.he_mcs[i];

			if (he_mcs == 0xFFFF)
				continue;

			nss = 0;
			max_mcs = 0;
			for (j = 0; j < sizeof(he_mcs) * 8; j+=2) {
				bits = (he_mcs >> j) & 0x3;
				if (bits == 0x3)
					continue;

				libwifi_dbg("%s nss%d %s\n", he_get_table(i), j/2, he_get_mcs(bits));
				switch (bits) {
				case 0:
					max_mcs = 7;
					break;
				case 1:
					max_mcs = 9;
					break;
				case 2:
					max_mcs = 11;
					break;
				default:
					max_mcs = 0;
					break;
				}

				switch (i) {
				case 0:
				case 1:
					bw = 80;
					break;
				case 2:
				case 3:
					bw = 160;
					break;
				/* TODO handle 80p80 */
				case 4:
				case 5:
					bw = 160;
					break;
				default:
					bw = 0;
				}

				nss++;
			}

			libwifi_dbg("bw %d nss %d max_mcs %d\n", bw, nss, max_mcs);
			if (max_mcs) {
				wsta->rate.rate = wifi_mcs2rate(max_mcs, bw, nss, sgi);
				wsta->rate.m.mcs = max_mcs;
				wsta->maxrate = wsta->rate.rate;
			}
		}
	}

	/* Set current nss/bw */
	wsta->rate.m.nss = nss;
	wsta->rate.m.sgi = sgi;

	switch (sta->link_bw) {
	case 1:
		wsta->rate.m.bw = BW20;
		break;
	case 2:
		wsta->rate.m.bw = BW40;
		break;
	case 3:
		wsta->rate.m.bw = BW80;
		break;
	case 4:
		wsta->rate.m.bw = BW160;
		break;
	default:
		wsta->rate.m.bw = BW_AUTO;
		break;
	}

	if (swap)
		sta->wnm_cap = BCMSWAP32(sta->wnm_cap);
	if (!!(sta->wnm_cap & 0x1))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_11V_BSS_TRANS);

	wsta->caps.valid |= WIFI_CAP_RM_VALID;
	if (sta->rrm_capabilities[0] & (1 << DOT11_RRM_CAP_LINK))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_LINK);
	if (sta->rrm_capabilities[0] & (1 << DOT11_RRM_CAP_NEIGHBOR_REPORT))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_NBR_REPORT);
	if (sta->rrm_capabilities[0] & (1 << DOT11_RRM_CAP_BCN_PASSIVE))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_BCN_PASSIVE);
	if (sta->rrm_capabilities[0] & (1 << DOT11_RRM_CAP_BCN_ACTIVE))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_BCN_ACTIVE);
	if (sta->rrm_capabilities[0] & (1 << DOT11_RRM_CAP_BCN_TABLE))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_BCN_TABLE);
	if (sta->rrm_capabilities[3] & (1 << (DOT11_RRM_CAP_RCPI % 8)))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_RCPI);
	if (sta->rrm_capabilities[3] & (1 << (DOT11_RRM_CAP_RSNI % 8)))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_RM_RSNI);
	if (sta->rrm_capabilities[3] & (1 << (DOT11_RRM_CAP_BSSAAD % 8)))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_BSS_AVG_ACCESS_DELAY);
	if (sta->rrm_capabilities[4] & (1 << (DOT11_RRM_CAP_BSSAAC % 8)))
		wifi_cap_set(wsta->cbitmap, WIFI_CAP_BSS_AVAIL_ADMISSION);

	/* status flags */

	/* libwifi_dbg("flags: 0x%08x\n", sta->flags); */
	if (!!(sta->flags & WL_STA_WME))
		wifi_status_set(wsta->sbitmap, WIFI_STATUS_WMM);

	if (!!(sta->flags & WL_STA_PS))
		wifi_status_set(wsta->sbitmap, WIFI_STATUS_PS);

	wsta->conn_time = sta->in;
	wsta->idle_time = sta->idle;
	wsta->tx_rate.rate = sta->tx_rate / 1000;  /* in Mbps */
	wsta->rx_rate.rate = sta->rx_rate / 1000;  /* in Mbps */

	if (wsta->maxrate == 0) {
		if (sta->rateset_adv.count < WL_MAXRATES_IN_SET) {
			wsta->maxrate =
				(sta->rateset_adv.rates[sta->rateset_adv.count] & 0x7f) * 2;
		}
	}

	for (i = 0; i < 4; i++) {
		wsta->rssi[i] = sta->rssi[i];
		wsta->noise[i] = sta->nf[i];
	}

	if (sta->flags & WL_STA_SCBSTATS) {
		wsta->stats.tx_pkts = sta->tx_tot_pkts;
		wsta->stats.rx_pkts = sta->rx_tot_pkts;
		wsta->stats.tx_bytes = sta->tx_tot_bytes;
		wsta->stats.rx_bytes = sta->rx_tot_bytes;
		wsta->stats.tx_fail_pkts = sta->tx_failures;
		wsta->stats.rx_fail_pkts = sta->rx_decrypt_failures; /* FIXME */
		wsta->stats.tx_retry_pkts = sta->tx_pkts_retried;
		//wsta->stats.tx_drop_pkts = sta->tx_failures;
		//wsta->stats.rx_drop_pkts = sta->rx_decrypt_failures;
	}


	/* to get airtime, instant thput etc. */
	bcm_get_sta_info_airtime(ifname, addr, wsta);

	return 0;
}

int bcmwl_iface_chan_switch(const char *ifname, struct chan_switch_param *param)
{
	char buf[1024] = {0};
	int bandwidth;
	int channel;
	int count;
	int status;

	libwifi_dbg("[%s] %s called\n", ifname, __func__);

	count = param->count;
	channel = f2c(param->freq);
	bandwidth = param->bandwidth;

	status = Cmd(buf, sizeof(buf), "wl -i %s csa 0 %d %d/%d", ifname, count, channel, bandwidth);
	WARN_ON(status);

	return status;
}

int bcmwl_iface_link_measure(const char *ifname, uint8_t *sta)
{
	char buf[1024] = {0};
	int ret;


	libwifi_dbg("[%s] %s called\n", ifname, __func__);

	memcpy(buf, sta, 6);
	ret = wl_iovar_set(ifname, "rrm_lm_req", sta, 6, buf, sizeof(buf));
	if (ret)
		libwifi_dbg("[%s] %s Failed\n", ifname, __func__);

	return ret;
}

int bcmwl_iface_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta_config *cfg)
{
	runCmd("wlctl -i %s sta_monitor %s " MACSTR,
	       ifname, cfg->enable ? "add" : "del", MAC2STR(sta));

	if (cfg->enable)
		runCmd("wlctl -i %s sta_monitor enable", ifname);

	return 0;
}

int bcmwl_iface_get_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta *monsta)
{
	char buf[256] = {0};
	char *pos = NULL;

	memset(monsta, 0, sizeof(*monsta));
	memcpy(monsta->macaddr, sta, 6);

	/* get monitored data */
	chrCmd(buf, sizeof(buf),
		"wlctl -i %s sta_monitor stats " MACSTR,
		ifname, MAC2STR(sta));
	if (buf[0]) {
		pos = strstr(buf, "RSSI :");
		if (pos) {
			pos += 6;
			monsta->rssi[0] = strtol(pos, NULL, 10);
			monsta->rssi[1] = 0;
			monsta->rssi[2] = 0;
			monsta->rssi[3] = 0;
		}
		monsta->last_seen = 0;
	}

	return 0;
}

int bcmwl_iface_add_vendor_ie(const char *ifname, struct vendor_iereq *req)
{
	char iebuf[512] = { 0 };
	size_t ielen;

	ielen = req->ie.ie_hdr.len;
	if (ielen > 254)
		return -1;

	btostr(req->ie.data, ielen - 3, iebuf);
	runCmd("wlctl -i %s add_ie %d %d %02x:%02x:%02x %s",
	       ifname, 15, ielen,
	       req->ie.oui[0], req->ie.oui[1], req->ie.oui[2],
	       iebuf);

	return 0;
}

int bcmwl_iface_del_vendor_ie(const char *ifname, struct vendor_iereq *req)
{
	char iebuf[512] = { 0 };
	size_t ielen;

	ielen = req->ie.ie_hdr.len;
	if (ielen > 254)
		return -1;

	btostr(req->ie.data, ielen - 3, iebuf);
	runCmd("wlctl -i %s del_ie %d %d %02x:%02x:%02x %s",
	       ifname, 15, ielen,
	       req->ie.oui[0], req->ie.oui[1], req->ie.oui[2],
	       iebuf);

	return 0;
}
