/*
 * wifi.c - API glue
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <openssl/rand.h>

#include "util.h"
#include "wifi.h"

extern const struct wifi_driver *wifi_drivers[];
extern uint32_t num_wifi_drivers;

const struct wifi_driver *get_wifi_driver(const char *ifname)
{
	int i;

	if (!ifname || ifname[0] == '\0')
		return NULL;

	for (i = 0; i < num_wifi_drivers; i++) {
		char *token = NULL, *ifpx, *tmp = NULL;
		char drv[16] = {0};

		strncpy(drv, wifi_drivers[i]->name, 15);
		for (ifpx = drv; ; ifpx = NULL) {
			token = strtok_r(ifpx, ",", &tmp);
			if (token == NULL)
				break;

			if (strstr(ifname, token))
				return wifi_drivers[i];
		}


		if (strstr(ifname, drv))
			return wifi_drivers[i];
	}

	return NULL;
}


int wifi_driver_info(const char *name, struct wifi_metainfo *info)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->info)
		return drv->info(name, info);

	return -ENOTSUP;
}

int wifi_scan(const char *ifname, struct scan_param *p)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->scan)
		return drv->scan(ifname, p);

	return -ENOTSUP;
}

int wifi_get_scan_results(const char *ifname, struct wifi_bss *bsss, int *nr)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_scan_results)
		return drv->get_scan_results(ifname, bsss, nr);

	return -ENOTSUP;
}

int wifi_get_bss_scan_result(const char *ifname, uint8_t *bssid,
						struct wifi_bss_detail *b)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_bss_scan_result)
		return drv->get_bss_scan_result(ifname, bssid, b);

	return -ENOTSUP;
}

#if 0
int wifi_get_device_id(char *ifname, int *id)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_device_id)
		return drv->get_device_id(ifname, id);

	return -ENOTSUP;
}
#endif

int wifi_get_ifstatus(const char *ifname, ifstatus_t *f)
{
	return get_ifstatus(ifname, f);
}

int wifi_get_ifoperstatus(const char *ifname, ifopstatus_t *opstatus)
{
	return if_getoperstate(ifname, opstatus);
}

int wifi_get_caps(const char *ifname, struct wifi_caps *caps)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_caps)
		return drv->iface.get_caps(ifname, caps);

	return -ENOTSUP;
}

int wifi_radio_get_caps(const char *name, struct wifi_caps *caps)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->radio.get_caps)
		return drv->radio.get_caps(name, caps);

	return -ENOTSUP;
}

int wifi_get_bssid(const char *ifname, uint8_t *bssid)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_bssid)
		return drv->get_bssid(ifname, bssid);

	return -ENOTSUP;
}

int wifi_get_ssid(const char *ifname, char *ssid)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_ssid)
		return drv->get_ssid(ifname, ssid);

	return -ENOTSUP;
}

int wifi_get_channel(const char *ifname, uint32_t *channel, enum wifi_bw *bw)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_channel)
		return drv->get_channel(ifname, channel, bw);

	return -ENOTSUP;
}

int wifi_vendor_cmd(const char *ifname, uint32_t vid, uint32_t subcmd,
			uint8_t *in, int ilen, uint8_t *out, int *olen)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->vendor_cmd)
		return drv->vendor_cmd(ifname, vid, subcmd, in, ilen, out, olen);

	return -ENOTSUP;
}


int wifi_subscribe_frame(const char *ifname, uint8_t type, uint8_t stype)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.subscribe_frame)
		return drv->iface.subscribe_frame(ifname, type, stype);

	return -ENOTSUP;
}

int wifi_unsubscribe_frame(const char *ifname, uint8_t type, uint8_t stype)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.unsubscribe_frame)
		return drv->iface.unsubscribe_frame(ifname, type, stype);

	return -ENOTSUP;
}

int wifi_set_4addr(const char *ifname, bool enable)
{

	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->set_4addr)
		return drv->set_4addr(ifname, enable);

	return -ENOTSUP;
}

int wifi_set_vlan(const char *ifname, struct vlan_param vlan)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->set_vlan)
		return drv->set_vlan(ifname, vlan);

	return -ENOTSUP;
}

int wifi_link_measure(const char *ifname, uint8_t *sta)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->link_measure)
		return drv->link_measure(ifname, sta);

	return -ENOTSUP;
}

int wifi_set_channel(const char *ifname, uint32_t channel, enum wifi_bw bw)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->set_channel)
		return drv->set_channel(ifname, channel, bw);

	return -ENOTSUP;
}

int wifi_chan_switch(const char *ifname, struct chan_switch_param *param)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->chan_switch)
		return drv->chan_switch(ifname, param);

	return -ENOTSUP;
}
int wifi_get_supp_channels(const char *ifname, uint32_t *chlist, int *num,
					const char *cc, enum wifi_band band,
					enum wifi_bw bw)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_supp_channels)
		return drv->get_supp_channels(ifname, chlist, num,
						cc, band, bw);

	return wifi_get_valid_channels(ifname, band, bw, cc, chlist, num);
}

int wifi_get_oper_stds(const char *name, uint8_t *std)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_oper_stds)
		return drv->get_oper_stds(name, std);

	return -ENOTSUP;
}

int wifi_get_country(const char *name, char *alpha2)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_country)
		return drv->get_country(name, alpha2);

	return -ENOTSUP;
}

int wifi_get_oper_band(const char *name, enum wifi_band *band)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_oper_band)
		return drv->get_oper_band(name, band);

	return -ENOTSUP;
}

int wifi_get_supp_band(const char *name, uint32_t *bands)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_supp_band)
		return drv->get_supp_band(name, bands);

	return -ENOTSUP;
}

int wifi_get_mode(const char *ifname, enum wifi_mode *mode)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_mode)
		return drv->iface.get_mode(ifname, mode);

	return -ENOTSUP;
}

int wifi_get_security(const char *ifname, uint32_t *auth, uint32_t *enc)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_security)
		return drv->iface.get_security(ifname, auth, enc);

	return -ENOTSUP;
}

int wifi_get_maxrate(const char *ifname, unsigned long *rate)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_maxrate)
		return drv->get_maxrate(ifname, rate);

	return -ENOTSUP;
}

int wifi_get_supp_opclass(const char *name, int *num_opclass,
						struct wifi_opclass *o)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_supp_opclass)
		return drv->get_supp_opclass(name, num_opclass, o);

	return wifi_get_supported_opclass(name, num_opclass, o);
}

int wifi_get_curr_opclass(const char *name, struct wifi_opclass *o)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_curr_opclass)
		return drv->get_curr_opclass(name, o);

	return wifi_get_opclass(name, o);
}

int wifi_get_bandwidth(const char *name, enum wifi_bw *bw)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->get_bandwidth)
		return drv->get_bandwidth(name, bw);

	return -ENOTSUP;
}

int wifi_get_noise(const char *ifname, int *noise)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_noise)
		return drv->get_noise(ifname, noise);

	return -ENOTSUP;
}

int wifi_radio_get_stats(const char *ifname, struct wifi_radio_stats *s)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->radio.get_stats)
		return drv->radio.get_stats(ifname, s);

	return -ENOTSUP;
}

int wifi_radio_list(struct radio_entry *radio, int *num)
{
	const struct wifi_driver *drv;
	struct radio_entry *r;
	int cur_num;
	int n;
	int i;

	cur_num = 0;
	n = *num;
	r = radio;

        /*
         * Return complete list or fail. In case of fail upper
         * layer will fallback to UCI.
         */
	for (i = 0; i < num_wifi_drivers; i++) {
		drv = wifi_drivers[i];
		if (!drv)
			return -1;
		if (!drv->radio_list)
			return -1;

		if (drv->radio_list(r, &n))
			return -1;

		cur_num += n;
		if (cur_num >= *num)
			break;

		/* Calc left size */
		n = *num - cur_num;
		r = &radio[cur_num];
	}

	*num = cur_num;

	return 0;
}

int wifi_radio_info(const char *ifname, struct wifi_radio *info)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->radio.info)
		return drv->radio.info(ifname, info);

	return -ENOTSUP;
}

int wifi_radio_get_hwaddr(const char *ifname, uint8_t *hwaddr)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->radio.get_hwaddr)
		return drv->radio.get_hwaddr(ifname, hwaddr);

	return -ENOTSUP;
}

int wifi_ap_info(const char *ifname, struct wifi_ap *info)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.ap_info)
		return drv->iface.ap_info(ifname, info);

	return -ENOTSUP;
}

int wifi_ap_get_stats(const char *ifname, struct wifi_ap_stats *s)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_stats)
		return drv->iface.get_stats(ifname, s);

	return -ENOTSUP;
}

int wifi_get_beacon_ies(const char *ifname, uint8_t *ies, int *len)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_beacon_ies)
		return drv->get_beacon_ies(ifname, ies, len);

	return -ENOTSUP;
}

int wifi_acs(const char *ifname, struct acs_param *p)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->acs)
		return drv->acs(ifname, p);

	return -ENOTSUP;
}

int wifi_get_assoclist(const char *ifname, uint8_t *stas, int *num)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_assoclist)
		return drv->get_assoclist(ifname, stas, num);

	return -ENOTSUP;
}

int wifi_get_sta_info(const char *ifname, uint8_t *addr, struct wifi_sta *info)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_sta_info)
		return drv->get_sta_info(ifname, addr, info);

	return -ENOTSUP;
}

int wifi_radio_get_param(const char *name, const char *param, int *len, void *val)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->radio.get_param)
		return drv->radio.get_param(name, param, len, val);

	return -ENOTSUP;
}

int wifi_add_iface(const char *name, enum wifi_mode m, char *argv[])
{
	const struct wifi_driver *drv = get_wifi_driver(name);
	int ret = -ENOTSUP;

	if (drv && drv->add_iface)
		ret = drv->add_iface(name, m, argv);

	return ret;
}

int wifi_del_iface(const char *name, const char *ifname)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->del_iface)
		return drv->del_iface(name, ifname);

	return -ENOTSUP;
}

int wifi_list_iface(const char *name, struct iface_entry *iface, int *num)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->list_iface)
		return drv->list_iface(name, iface, num);

	return -ENOTSUP;
}

int wifi_channels_info(const char *name, struct chan_entry *channel, int *num)
{
	const struct wifi_driver *drv = get_wifi_driver(name);

	if (drv && drv->channels_info)
		return drv->channels_info(name, channel, num);

	return -ENOTSUP;
}

#if 0
int wifi_get_sta_stats(const char *ifname, uint8_t *addr, struct wifi_sta_stats *sts)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_sta_stats)
		return drv->get_sta_stats(ifname, addr, sts);

	return -1;
}
#endif

int wifi_disconnect_sta(const char *ifname, uint8_t *sta, uint16_t reason)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->disconnect_sta)
		return drv->disconnect_sta(ifname, sta, reason);

	return -ENOTSUP;
}

int wifi_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta_config *cfg)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->monitor_sta)
		return drv->monitor_sta(ifname, sta, cfg);

	return -ENOTSUP;
}

int wifi_get_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta *mon)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_monitor_sta)
		return drv->get_monitor_sta(ifname, sta, mon);

	return -ENOTSUP;
}

int wifi_get_monitor_stas(const char *ifname, struct wifi_monsta *stas, int *num_stas)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->get_monitor_stas)
		return drv->get_monitor_stas(ifname, stas, num_stas);

	return -ENOTSUP;

}

int wifi_restrict_sta(const char *ifname, uint8_t *sta, int enable)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->restrict_sta)
		return drv->restrict_sta(ifname, sta, enable);

	return -ENOTSUP;
}

int wifi_get_neighbor_list(const char *ifname, struct nbr *nbrs, int *nr)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_neighbor_list)
		return drv->iface.get_neighbor_list(ifname, nbrs, nr);

	return -ENOTSUP;
}

int wifi_add_neighbor(const char *ifname, struct nbr nbr)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.add_neighbor)
		return drv->iface.add_neighbor(ifname, nbr);

	return -ENOTSUP;
}

int wifi_del_neighbor(const char *ifname, unsigned char *bssid)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.del_neighbor)
		return drv->iface.del_neighbor(ifname, bssid);

	return -ENOTSUP;
}

int wifi_req_beacon_report(const char *ifname, unsigned char *da)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.req_beacon_report)
		return drv->iface.req_beacon_report(ifname, da);

	return -ENOTSUP;
}

int wifi_get_beacon_report(const char *ifname, unsigned char *sta,
				struct sta_nbr *snbrs, int *nr)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_beacon_report)
		return drv->iface.get_beacon_report(ifname, sta, snbrs, nr);

	return -ENOTSUP;
}

int wifi_req_bss_transition(const char *ifname, unsigned char *sta,
		int bsss_nr, unsigned char *bsss, unsigned int tmo)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.req_bss_transition)
		return drv->iface.req_bss_transition(ifname, sta, bsss_nr,
				bsss, tmo);

	return -ENOTSUP;
}

int wifi_req_btm(const char *ifname, uint8_t *sta, int bsss_nr,
				uint8_t *bsss, struct wifi_btmreq *b)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.req_btm)
		return drv->iface.req_btm(ifname, sta, bsss_nr, bsss, b);

	return -ENOTSUP;
}

int wifi_get_11rkeys(const char *ifname, unsigned char *sta,
			uint8_t *r1khid)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_11rkeys)
		return drv->iface.get_11rkeys(ifname, sta, r1khid);

	return -ENOTSUP;
}

int wifi_set_11rkeys(const char *ifname, struct fbt_keys *fk)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.set_11rkeys)
		return drv->iface.set_11rkeys(ifname, fk);

	return -ENOTSUP;
}

int wifi_add_vendor_ie(const char *ifname, struct vendor_iereq *req)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.add_vendor_ie)
		return drv->iface.add_vendor_ie(ifname, req);

	return -ENOTSUP;
}

int wifi_del_vendor_ie(const char *ifname, struct vendor_iereq *req)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.del_vendor_ie)
		return drv->iface.del_vendor_ie(ifname, req);

	return -ENOTSUP;
}

int wifi_get_vendor_ies(const char *ifname, struct vendor_ie *ies, int *num_ies)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_vendor_ies)
		return drv->iface.get_vendor_ies(ifname, ies, num_ies);

	return -ENOTSUP;
}

int wifi_sta_info(const char *ifname, struct wifi_sta *sta)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.sta_info)
		return drv->iface.sta_info(ifname, sta);

	return -ENOTSUP;
}

int wifi_sta_get_stats(const char *ifname, struct wifi_sta_stats *s)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.sta_get_stats)
		return drv->iface.sta_get_stats(ifname, s);

	return -ENOTSUP;
}

int wifi_sta_get_ap_info(const char *ifname, struct wifi_bss *info)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.sta_get_ap_info)
		return drv->iface.sta_get_ap_info(ifname, info);

	return -ENOTSUP;
}

int wifi_sta_disconnect_ap(const char *ifname, uint32_t reason)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.sta_disconnect_ap)
		return drv->iface.sta_disconnect_ap(ifname, reason);

	return -ENOTSUP;
}

int wifi_register_event(const char *ifname, struct event_struct *ev, void **handle)
{
	const struct wifi_driver *drv;
	int ret = -1;

	if (!ifname)
		return easy_register_event(ev, handle);

	drv = get_wifi_driver(ifname);
	if (drv) {
		if (drv->register_event)
			ret = drv->register_event(ifname, ev, handle);
		else
			ret = easy_register_event(ev, handle);
	}

	return ret;
}

int wifi_unregister_event(const char *ifname, void *handle)
{
	const struct wifi_driver *drv;
	int ret = -1;

	if (!ifname || ifname[0] == '\0')
		return easy_unregister_event(handle);

	drv = get_wifi_driver(ifname);
	if (drv) {
		if (drv->unregister_event)
			ret = drv->unregister_event(ifname, handle);
		else
			ret = easy_unregister_event(handle);
	}

	return ret;
}

int wifi_recv_event(const char *ifname, void *handle)
{
	const struct wifi_driver *drv;
	int ret = -1;

	if (!ifname || ifname[0] == '\0')
		return easy_recv_event(handle);

	drv = get_wifi_driver(ifname);
	if (drv) {
		if (drv->recv_event)
			ret = drv->recv_event(ifname, handle);
		else
			ret = easy_recv_event(handle);
	}

	return ret;
}

int wifi_start_wps(const char *ifname, struct wps_param wps)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.start_wps)
		return drv->iface.start_wps(ifname, wps);

	return -ENOTSUP;
}

int wifi_stop_wps(const char *ifname)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.stop_wps)
		return drv->iface.stop_wps(ifname);

	return -ENOTSUP;
}

int wifi_get_wps_status(const char *ifname, enum wps_status *s)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_wps_status)
		return drv->iface.get_wps_status(ifname, s);

	return -ENOTSUP;
}

int wifi_get_wps_pin(const char *ifname, unsigned long *pin)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_wps_pin)
		return drv->iface.get_wps_pin(ifname, pin);

	return -ENOTSUP;
}

int wifi_set_wps_pin(const char *ifname, unsigned long pin)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.set_wps_pin)
		return drv->iface.set_wps_pin(ifname, pin);

	return -ENOTSUP;
}

int wifi_get_wps_device_info(const char *ifname, struct wps_device *info)
{
	const struct wifi_driver *drv = get_wifi_driver(ifname);

	if (drv && drv->iface.get_wps_device_info)
		return drv->iface.get_wps_device_info(ifname, info);

	return -ENOTSUP;
}

int wifi_generate_wps_pin(unsigned long *pin)
{
	unsigned long PIN;
	unsigned long int accum = 0;
	int digit;
	char local_devPwd[32];
	union {
		unsigned char rand_bytes[8];
		uint32_t rand_uint32[2];
	} rand_data;

	RAND_bytes(rand_data.rand_bytes, 8);
	sprintf(local_devPwd, "%08u", *(uint32_t *)rand_data.rand_uint32);
	local_devPwd[7] = '\0';
	PIN = strtoul(local_devPwd, NULL, 10);

	PIN *= 10;
	accum += 3 * ((PIN / 10000000) % 10);
	accum += 1 * ((PIN / 1000000) % 10);
	accum += 3 * ((PIN / 100000) % 10);
	accum += 1 * ((PIN / 10000) % 10);
	accum += 3 * ((PIN / 1000) % 10);
	accum += 1 * ((PIN / 100) % 10);
	accum += 3 * ((PIN / 10) % 10);

	digit = (accum % 10);
	accum = (10 - digit) % 10;

	PIN += accum;
	*pin = PIN;

	return 0;
}

int wifi_is_wps_pin_valid(unsigned long pin)
{
	unsigned long int accum = 0;
	unsigned int len;
	int valid = 0;
	char pinstr[16] = {0};

	sprintf(pinstr, "%lu", pin);
	len = strlen(pinstr);
	if (len == 8) {
		accum += 3 * ((pin / 10000000) % 10);
		accum += 1 * ((pin / 1000000) % 10);
		accum += 3 * ((pin / 100000) % 10);
		accum += 1 * ((pin / 10000) % 10);
		accum += 3 * ((pin / 1000) % 10);
		accum += 1 * ((pin / 100) % 10);
		accum += 3 * ((pin / 10) % 10);
		accum += 1 * ((pin / 1) % 10);

		if ((accum % 10) == 0)
			valid = 1;
	} else if (len == 4) {
		valid = 1;
	}

	return valid;
}

/*
 * This is actually an ABI list. Do NOT change the order of the apis
 * in the list unless similar adjustment is made within the wifi_driver
 * operations structure.
 * It serves as a convenient list to see at a glance the APIs that
 * libwifi currently supports.
 */
const char *libwifi_apis[] = {
	"libwifi_get_apis",    /* always should be first */

	"wifi_driver_info",
	"wifi_radio_list",

	/*
	 * wifi radio apis
	 */
	"wifi_radio_info",
	"wifi_get_supp_band",
	"wifi_get_oper_band",
	"wifi_radio_get_caps",
	"wifi_get_supp_stds",
	"wifi_get_oper_stds",
	"wifi_get_country",
	"wifi_get_channel",
	"wifi_set_channel",  /* deprecate */
	"wifi_get_supp_channels",
	"wifi_get_oper_channels",
	"wifi_get_supp_opclass",
	"wifi_get_curr_opclass",
	"wifi_get_bandwidth",
	"wifi_get_maxrate",
	"wifi_get_basic_rates",
	"wifi_get_oper_rates",
	"wifi_get_supp_rates",
	"wifi_radio_get_stats",
	"wifi_scan",
	"wifi_get_scan_results",
	"wifi_get_bss_scan_result",
	"wifi_get_noise",
	"wifi_acs",
	"wifi_radio_get_param",
	"wifi_radio_set_param",
	"wifi_radio_get_hwaddr",
	"wifi_add_iface",
	"wifi_del_iface",
	"wifi_list_iface",
	"wifi_channels_info",


	/*
	 * wifi interface apis
	 */
	"wifi_start_wps",
	"wifi_stop_wps",
	"wifi_get_wps_status",
	"wifi_get_wps_pin",
	"wifi_set_wps_pin",
	"wifi_get_wps_device_info",
	"wifi_get_caps",
	"wifi_get_mode",
	"wifi_get_security",
	"wifi_add_vendor_ie",
	"wifi_del_vendor_ie",
	"wifi_get_vendor_ies",
	"wifi_get_param",
	"wifi_set_param",
	"wifi_vendor_cmd",
	"wifi_subscribe_frame",
	"wifi_unsubscribe_frame",
	"wifi_set_4addr",
	"wifi_set_vlan",
	"wifi_link_measure",

	"wifi_ap_info",
	"wifi_get_bssid",
	"wifi_get_ssid",
	"wifi_ap_get_stats",
	"wifi_get_beacon_ies",
	"wifi_get_assoclist",
	"wifi_get_sta_info",
	"wifi_get_sta_stats",
	"wifi_disconnect_sta",
	"wifi_restrict_sta",
	"wifi_monitor_sta",
	"wifi_get_monitor_sta",
	"wifi_get_monitor_stas",
	"wifi_add_neighbor",
	"wifi_del_neighbor",
	"wifi_get_neighbor_list",
	"wifi_req_beacon_report",
	"wifi_get_beacon_report",
	"wifi_req_bss_transition",
	"wifi_req_btm",
	"wifi_get_11rkeys",
	"wifi_set_11rkeys",
	"wifi_chan_switch",

	"wifi_sta_info",
	"wifi_sta_get_stats",
	"wifi_sta_get_ap_info",
	"wifi_sta_disconnect_ap",

	"wifi_register_event",
	"wifi_recv_event",
	"libwifi_get_version",   /* always the last api */
};

#define libwifi_apis_num  sizeof(libwifi_apis) / sizeof(libwifi_apis[0])

const char **libwifi_get_apis(char *name)
{
	return libwifi_apis;
}

const char *libwifi_get_version(void)
{
	return libwifi_version;
}

typedef int (*libwifi_api_t)(char *name, ...);

int libwifi_supports(const char *ifname, char *apiname)
{
	const struct wifi_driver *drv;
	libwifi_api_t find;
	int num_radio_apis;
	uint8_t *base;
	int i, idx;

	if (!ifname || !apiname) {
		errno = -EINVAL;
		return 0;
	}

	drv = get_wifi_driver(ifname);
	if (!drv) {
		errno = -EINVAL;
		return 0;
	}

	num_radio_apis = sizeof(drv->radio)/sizeof(libwifi_api_t);
	base = (uint8_t *)drv;

	for (i = 0; i < libwifi_apis_num; i++) {
		if (strncmp(libwifi_apis[i], apiname, strlen(libwifi_apis[i])))
			continue;

		if (i < 3)
			return 1;

		idx = i - 3;
		base += idx <= num_radio_apis ?
			offsetof(struct wifi_driver, radio) :
			offsetof(struct wifi_driver, iface);

		if (idx > num_radio_apis)
			idx -= num_radio_apis;

		find = *((libwifi_api_t *)(base +
				idx * sizeof(libwifi_api_t)));

		return (!find) ? 0 : 1;
	}

	errno = -ENOTSUP;
	return 0;
}
